from itertools import product

import numpy as np
import pytest

from PartSegCore_compiled_backend.sprawl_utils.euclidean_cython import calculate_euclidean


@pytest.fixture
def cube_data():
    data = np.zeros((10, 10, 10), dtype=np.uint8)
    data[3:-3, 3:-3, 3:-3] = 1
    return data


@pytest.fixture
def neigh():
    return np.array([x for x in product(range(-1, 2), repeat=3) if x != (0, 0, 0)], dtype=np.int8)


def test_calculate_euclidean(cube_data, neigh):
    res = calculate_euclidean(
        np.ones(cube_data.shape, dtype=np.uint8), cube_data, neigh, np.max(np.abs(neigh), axis=1).astype(np.float64)
    )
    expected = np.ones(cube_data.shape, dtype=np.uint8) * 3
    expected[1:-1, 1:-1, 1:-1] = 2
    expected[2:-2, 2:-2, 2:-2] = 1
    expected[3:-3, 3:-3, 3:-3] = 0
    assert np.all(res == expected)


def test_calculate_euclidean_mask(cube_data, neigh):
    mask = np.zeros(cube_data.shape, dtype=np.uint8)
    mask[1:-1, 1:-1, 1:-1] = 1
    res = calculate_euclidean(mask, cube_data, neigh, np.max(np.abs(neigh), axis=1).astype(np.float64))
    expected = np.zeros(cube_data.shape, dtype=np.float64)
    expected[:] = np.inf
    expected[1:-1, 1:-1, 1:-1] = 2
    expected[2:-2, 2:-2, 2:-2] = 1
    expected[3:-3, 3:-3, 3:-3] = 0
    assert np.all(res == expected)



import numpy as np
import numpy.testing as npt
import pytest

from PartSegCore_compiled_backend.napari_mapping import (
    map_array_parallel,
    map_array_sequential,
    zero_preserving_modulo_parallel,
    zero_preserving_modulo_sequential,
)


def _zero_preserving_modulo_numpy(values: np.ndarray, n: int, dtype: np.dtype, to_zero: int = 0) -> np.ndarray:
    res = ((values - 1) % n + 1).astype(dtype)
    res[values == to_zero] = 0
    return res


def _map_array(values: np.ndarray, map_dict: dict, dtype: np.dtype) -> np.ndarray:
    return np.array([map_dict.get(x, 0) for x in values.flatten()], dtype=dtype).reshape(values.shape)


DATA_LI = [
    np.arange(100, dtype=np.uint32),
    np.arange(100, dtype=np.int32),
    np.arange(100, dtype=np.uint64),
    np.arange(100, dtype=np.int64),
    np.arange(100, dtype=np.uint32).reshape((10, 10)),
]
DATA_IDS = ['uint32', 'int32', 'uint64', 'int64', 'uint32_2d']


@pytest.mark.parametrize('func', [zero_preserving_modulo_parallel, zero_preserving_modulo_sequential])
@pytest.mark.parametrize('data', DATA_LI, ids=DATA_IDS)
def test_zero_preserving_modulo(func, data):
    out = func(data, 10, 0)
    assert out.dtype == np.uint8
    npt.assert_array_equal(out, _zero_preserving_modulo_numpy(data, 10, np.uint8, 0))


@pytest.mark.parametrize('func', [zero_preserving_modulo_parallel, zero_preserving_modulo_sequential])
@pytest.mark.parametrize('background_num', [0, 1, 2, -1])
def test_background_label(func, background_num):
    data = np.zeros((10, 10), dtype=np.int32)
    data[1:-1, 1:-1] = 1
    data[2:-2, 2:-2] = 2
    data[4:-4, 4:-4] = -1

    res = func(data, 49, background_num)
    np.testing.assert_array_equal(res == 0, data == background_num)
    np.testing.assert_array_equal(res != 0, data != background_num)


@pytest.mark.parametrize('func', [map_array_parallel, map_array_sequential])
@pytest.mark.parametrize('data', DATA_LI, ids=DATA_IDS)
def test_map_array(func, data):
    map_dkt = {1: 1, 2: 2, 3: 1, 4: 3, 5: 2}
    out = func(data, map_dkt, 0)
    assert out.dtype == np.uint8
    npt.assert_array_equal(out, _map_array(data, map_dkt, np.uint8))


@pytest.mark.parametrize(('num', 'dtype'), [(40, np.uint8), (1000, np.uint16)])
@pytest.mark.parametrize('func', [zero_preserving_modulo_parallel, zero_preserving_modulo_sequential])
def test_cast_labels_to_minimum_type_auto(num: int, dtype, monkeypatch, func):
    data = np.zeros(3, dtype=np.uint32)
    data[1] = 10
    data[2] = 10**6 + 5
    cast_arr = func(data, num, 0)
    assert cast_arr.dtype == dtype
    assert cast_arr[0] == 0
    assert cast_arr[1] == 10
    assert cast_arr[2] == 10**6 % num + 5



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True
# cython: language_level=3

import numpy as np

cimport numpy as np
from libcpp.vector cimport vector
from numpy cimport float64_t, int8_t, uint8_t

from .distance_utils cimport Point, Size, component_types, my_queue

ctypedef fused object_area_types:
    float64_t
    uint8_t


cdef void put_borders_in_queue(my_queue[Point] & current_points,
                                      np.ndarray[uint8_t, ndim=3] base_object,
                                      np.ndarray[int8_t, ndim=2] neighbourhood) except *:
    cdef Size x_size, y_size, z_size, x, y, z, xx, yy, zz
    cdef Point p, p1
    cdef char neigh_length = neighbourhood.shape[0]
    z_size = base_object.shape[0]
    y_size = base_object.shape[1]
    x_size = base_object.shape[2]
    for z in range(0, z_size):
        for y in range(0, y_size):
            for x in range (0, x_size):
                if base_object[z,y,x] > 0:
                    for neigh_it in range(neigh_length):
                        zz = z+neighbourhood[neigh_it, 0]
                        yy = y+neighbourhood[neigh_it, 1]
                        xx = x+neighbourhood[neigh_it, 2]
                        if xx == -1 or xx == x_size or yy == -1 or yy == y_size or zz == -1 or zz == z_size:
                            continue
                        if base_object[zz, yy, xx] == 0:
                            p.z = z
                            p.y = y
                            p.x = x
                            current_points.push(p)
                            break


cdef vector[my_queue[Point]] create_borders_queues(np.ndarray[component_types, ndim=3] base_object,
                                      np.ndarray[int8_t, ndim=2] neighbourhood, component_types components_num)  except *:
    cdef Size x_size, y_size, z_size, x, y, z, xx, yy, zz
    cdef Point p, p1
    cdef char neigh_length = neighbourhood.shape[0]
    cdef size_t point_size = sizeof(my_queue[Point])
    cdef vector[my_queue[Point]] result = vector[my_queue[Point]](components_num+1)
    cdef component_types index
    print(components_num, point_size)
    z_size = base_object.shape[0]
    y_size = base_object.shape[1]
    x_size = base_object.shape[2]
    for z in range(1, z_size-1):
        for y in range(1, y_size-1):
            for x in range (1, x_size-1):
                if base_object[z,y,x] > 0:
                    index = base_object[z,y,x]
                    for neigh_it in range(neigh_length):
                        zz = z+neighbourhood[neigh_it, 0]
                        yy = y+neighbourhood[neigh_it, 1]
                        xx = x+neighbourhood[neigh_it, 2]
                        if xx == -1 or xx == x_size or xx == -1 or yy == y_size or zz == -1 or zz == z_size:
                            continue
                        if base_object[zz, yy, xx] != index:
                            p.z = z
                            p.y = y
                            p.x = x
                            result[index].push(p)
    return result



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True
# cython: language_level=3

from __future__ import print_function

from numpy cimport float64_t, int8_t, uint8_t

from .distance_utils cimport Point

include "put_borders_in_queue.pyx"

cpdef enum MuType:
    reversed_mu = 1
    reflection_mu = 2
    two_object_mu = 3

cdef inline double calculate_two_object_mu(float64_t pixel_val, float64_t lower_bound, float64_t upper_bound,
                                           float64_t lower_mid_bound, float64_t upper_mid_bound):
    cdef double mu = (pixel_val - lower_bound) / (upper_bound - lower_bound);
    if (lower_bound - lower_mid_bound) > 0 and (pixel_val >= lower_mid_bound) and pixel_val <= lower_bound:
        mu = (pixel_val - lower_mid_bound) / (lower_bound - lower_mid_bound)
    elif (upper_bound - lower_bound) > 0 and lower_bound < pixel_val <= upper_bound:
        mu = (upper_bound - pixel_val) / (upper_bound - lower_bound)
    if mu > 1:
        mu = 1
    if mu < 0:
        mu = 0
    return mu

cdef inline double calculate_mu(float64_t pixel_val, float64_t lower_bound, float64_t upper_bound,
                                const MuType flag):
    cdef double mu = (pixel_val - lower_bound) / (upper_bound - lower_bound);
    if flag == MuType.reversed_mu:
        mu = 1 - mu
    if flag == MuType.reflection_mu and mu < 0.5:
        mu = 1 - mu
    if mu > 1:
        mu = 1
    if mu < 0:
        mu = 0
    return mu


def calculate_mu_array(np.ndarray[float64_t, ndim=3] object_area, float64_t lower_bound, float64_t upper_bound,
                       MuType mu_type):
    cdef Size x_size, y_size, z_size, x, y, z, xx, yy, zz
    z_size = object_area.shape[0]
    y_size = object_area.shape[1]
    x_size = object_area.shape[2]
    cdef np.ndarray[float64_t, ndim=3] mu_array = np.zeros((z_size, y_size, x_size), dtype=np.float64)
    for z in range(1, z_size-1):
        for y in range(1, y_size-1):
            for x in range (1, x_size-1):
                mu_array[z, y, x] = calculate_mu(object_area[z,y,x], lower_bound, upper_bound, mu_type)
    return mu_array


def fuzzy_distance(np.ndarray[float64_t, ndim=3] object_area, np.ndarray[uint8_t, ndim=3] base_object,
                   np.ndarray[int8_t, ndim=2] neighbourhood, np.ndarray[float64_t, ndim=1] distance,
                   np.ndarray[float64_t, ndim=3] mu_array, distance_cache=None):
    cdef np.ndarray[uint8_t, ndim=3] consumed_area = np.copy(base_object)
    cdef np.ndarray[float64_t, ndim=3] result
    cdef Size x_size, y_size, z_size, array_pos, x, y, z, xx, yy, zz
    cdef Py_ssize_t count = 0, count2 = 0
    cdef char neigh_length = neighbourhood.shape[0]
    cdef int neigh_it
    cdef my_queue[Point] current_points, new_points
    cdef Point p, p1
    cdef double point_mu, neigh_mu, mu_diff
    cdef float64_t object_value, object_neighbourhood_value
    z_size = object_area.shape[0]
    y_size = object_area.shape[1]
    x_size = object_area.shape[2]
    result = np.zeros((z_size, y_size, x_size), dtype=np.float64)
    result[base_object == 0] = np.inf
    put_borders_in_queue(current_points, base_object, neighbourhood)
    while not current_points.empty():
        p = current_points.front()
        current_points.pop()
        count += 1
        """if consumed_area[p.z, p.y, p.x] > 0:
            continue"""
        consumed_area[p.z, p.y, p.x] = 0
        # object_value = object_area[p.z, p.y, p.x]
        point_mu = mu_array[p.z, p.y, p.x]
        for neigh_it in range(neigh_length):
            z = p.z + neighbourhood[neigh_it, 0]
            y = p.y + neighbourhood[neigh_it, 1]
            x = p.x + neighbourhood[neigh_it, 2]
            if x < 0 or y < 0 or z < 0 or x >= x_size or y >= y_size or z >= z_size:
                continue
            if object_area[z, y, x] == 0:
                continue
            # object_neighbourhood_value = object_area[z, y, x]
            neigh_mu = mu_array[z, y, x]
            mu_diff = (point_mu + neigh_mu) * distance[neigh_it] #  abs(object_value - object_neighbourhood_value)
            if result[z, y, x] > result[p.z, p.y, p.x] + mu_diff:
                result[z, y, x] = result[p.z, p.y, p.x] + mu_diff
                if consumed_area[z, y, x] == 0:
                    consumed_area[z, y, x] = 1
                    p1.z = z
                    p1.y = y
                    p1.x = x
                    current_points.push(p1)
    # print("total_steps: " +str(count), " and ", str(count2))
    return result



from __future__ import annotations

import numpy as np

from PartSegCore_compiled_backend._napari_mapping import (
    _map_array_par,
    _map_array_seq,
    _zero_preserving_modulo_par,
    _zero_preserving_modulo_seq,
)


def _allocate_output(data: np.ndarray, num_values: int) -> np.ndarray:
    if num_values < 256:
        dtype = np.uint8
    elif num_values < 65536:
        dtype = np.uint16
    else:
        dtype = np.float32
    return np.empty_like(data, dtype=dtype)


def zero_preserving_modulo_sequential(
    data: np.ndarray, modulo_factor: int, to_zero: int, out: np.ndarray | None = None
) -> np.ndarray:
    """
    Modulo plus one operation performed on values different than to_zero.
    `(n % modulo_factor) + 1 if n != to_zero else 0`

    Perform operation on each element of array sequentially.

    Parameters
    ----------
    data: np.ndarray
        data to be modified
    modulo_factor: int
        modulo factor
    to_zero: int
        value to be set to zero
    out: Optional[np.ndarray]
        output array
    """
    original_shape = data.shape
    data = data.reshape(-1)
    if out is None:
        out = _allocate_output(data, modulo_factor)
    _zero_preserving_modulo_seq(data, modulo_factor, to_zero, out)
    return out.reshape(original_shape)


def zero_preserving_modulo_parallel(
    data: np.ndarray, modulo_factor: int, to_zero: int, out: np.ndarray | None = None
) -> np.ndarray:
    """
    Modulo plus one operation performed on values different than to_zero.
    `(n % modulo_factor) + 1 if n != to_zero else 0`

    Perform operation on each element of array in parallel using openmp.

    Parameters
    ----------
    data: np.ndarray
        data to be modified
    modulo_factor: int
        modulo factor
    to_zero: int
        value to be set to zero
    out: Optional[np.ndarray]
        output array
    """
    original_shape = data.shape
    data = data.reshape(-1)
    if out is None:
        out = _allocate_output(data, modulo_factor)
    _zero_preserving_modulo_par(data, modulo_factor, to_zero, out)
    return out.reshape(original_shape)


def map_array_sequential(
    data: np.ndarray, mapping: dict, default: int = 0, out: np.ndarray | None = None
) -> np.ndarray:
    """
    Map values from data to values from mapping.

    Parameters
    ----------
    data: np.ndarray
        data to be modified
    mapping: dict
        dict with mapping information. Keys are values from data, values are values from output array.
        If None is provided then its value is set to default.
    default: int
        default value for not mapped values
    out: Optional[np.ndarray]
        output array
    """
    original_shape = data.shape
    data = data.reshape(-1)
    if out is None:
        out = _allocate_output(data, max(x for x in mapping.values() if x is not None) + 1)
    _map_array_seq(data, mapping, out, out.dtype.type(default))
    return out.reshape(original_shape)


def map_array_parallel(data: np.ndarray, mapping: dict, default: int = 0, out: np.ndarray | None = None) -> np.ndarray:
    """
    Map values from data to values from mapping.

    Parameters
    ----------
    data: np.ndarray
        data to be modified
    mapping: dict
        dict with mapping information. Keys are values from data, values are values from output array.
        If None is provided then its value is set to default.
    default: int
        default value for not mapped values
    out: Optional[np.ndarray]
        output array
    """
    original_shape = data.shape
    data = data.reshape(-1)
    if out is None:
        out = _allocate_output(data, max(x for x in mapping.values() if x is not None) + 1)
    _map_array_par(data, mapping, out, out.dtype.type(default))
    return out.reshape(original_shape)



import numpy as np

def calculate_borders(labels: np.ndarray, border_thick: int, per_layer: bool = True) -> np.ndarray: ...
def calculate_borders2d(labels: np.ndarray, border_thick: int) -> np.ndarray: ...
def color_grayscale(
    cmap: np.ndarray, image: np.ndarray, min_val: float, max_val: float, single_channel: int = -1
) -> np.ndarray: ...
def add_labels(
    image: np.ndarray,
    labels: np.ndarray,
    overlay: float,
    only_border: int,
    border_thick: int,
    use_labels: np.ndarray,
    label_colors: np.ndarray,
) -> np.ndarray: ...



from enum import Enum

import numpy as np

class MuType(Enum):
    base_mu = 1
    reflection_mu = 2
    two_object_mu = 3

def calculate_mu(
    image: np.ndarray,
    lower_bound: float,
    upper_bound: float,
    type_: MuType,
    mask: np.ndarray | None = None,
    lower_mid_bound: float = 0,
    upper_mid_bound: float = 0,
): ...

class PyMSO:
    def set_image(
        self,
        image: np.ndarray,
        image_types: float,
        lower_bound: float,
        upper_bound: float,
        type_: MuType,
        mask: np.ndarray | None = None,
        lower_mid_bound: float = 0,
        upper_mid_bound: float = 0,
    ): ...
    def set_mu_array(self, mu: np.ndarray): ...
    def set_components(self, components: np.ndarray, component_num: int | None = None): ...
    def set_neighbourhood(self, neighbourhood: np.ndarray, distances: np.ndarray): ...
    def calculate_FDT(self) -> np.ndarray: ...
    def optimum_erosion_calculate(
        self, fdt_array: np.ndarray, components_arr: np.ndarray, sprawl_area: np.ndarray
    ) -> np.ndarray: ...
    def constrained_dilation(
        self, fdt_array: np.ndarray, components_arr: np.ndarray, sprawl_area: np.ndarray
    ) -> np.ndarray: ...
    def run_MSO(self, step_limits: int = 1, count_steps_factor: int = 3): ...
    def steps_done(self) -> int: ...
    def set_components_num(self, num: int): ...
    def get_result_catted(self) -> np.ndarray: ...
    def get_fdt(self) -> np.ndarray: ...
    def set_use_background(self, use: bool): ...

def calculate_mu_mid(image: np.ndarray, lower_bound: float, mid_point: float, upper_bound: float): ...



import numpy as np

def calculate_euclidean(
    object_area: np.ndarray,
    base_object: np.ndarray,
    neighbourhood: np.ndarray,
    distance: np.ndarray,
    distance_cache: np.ndarray | None = None,
) -> np.ndarray: ...



# distutils: define_macros=CYTHON_TRACE_NOGIL=1
# cython: language_level=3, boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, embedsignature=True
"""
This module contains the cython implementation of the ``calc_bounds`` function.

currently supported dimensions: 2, 3, 4, 5
currently supported label types: uint8, uint16, uint32
"""


from typing import Optional
import numpy as np

cimport numpy as cnp

ctypedef fused label_types:
    cnp.uint8_t
    cnp.uint16_t
    cnp.uint32_t

def calc_bounds(labels: np.ndarray, components_num: Optional[int]=None):
    """
    Calculate the bounds of each component.
    wrapper around ``calc_boundsX`` function for different number of dimensions

    Parameters
    ----------
    labels : ndarray
        The labels of the components.
    components_num : int, optional
        The number of components.

    Returns
    -------
    bounds : (ndarray, ndarray)
        The bounds of each component.
    """
    if components_num is None:
        components_num = int(np.max(labels))
    return {
        2: calc_bounds2,
        3: calc_bounds3,
        4: calc_bounds4,
        5: calc_bounds5,
    }[labels.ndim](labels, components_num)


def calc_bounds5(cnp.ndarray[label_types, ndim=5] labels, Py_ssize_t components_num):
    """
    Calculate the bounds of the specified labels in a 5-dimensional array.

    Parameters
    ----------
    labels: numpy.ndarray[label_types, ndim=5]
        The 5-dimensional array containing the labels.
    components_num: Py_ssize_t
        The number of components to calculate the bounds for.

    Returns
    -------
    bounds: tuple[numpy.ndarray[cnp.int16_t, ndim=2], numpy.ndarray[cnp.int16_t, ndim=2]]
        The bounds of each component.
    """
    cdef Py_ssize_t x, y, z, t, s
    cdef Py_ssize_t x_max = labels.shape[4]
    cdef Py_ssize_t y_max = labels.shape[3]
    cdef Py_ssize_t z_max = labels.shape[2]
    cdef Py_ssize_t t_max = labels.shape[1]
    cdef Py_ssize_t s_max = labels.shape[0]
    cdef label_types label_val
    cdef cnp.ndarray[cnp.int16_t, ndim=2] min_bound = np.full((components_num + 1, 5), max(x_max, y_max, z_max, t_max, s_max) + 5, dtype=np.int16)
    cdef cnp.ndarray[cnp.int16_t, ndim=2] max_bound = np.full((components_num + 1, 5), -1, dtype=np.int16)

    for s in range(0, s_max):
        for t in range(0, t_max):
            for z in range(0, z_max):
                for y in range(0,y_max):
                    for x in range(0,x_max):
                        label_val = labels[s, t, z, y, x]
                        if label_val == 0:
                            continue
                        min_bound[label_val, 0] = min(min_bound[label_val, 0], s)
                        min_bound[label_val, 1] = min(min_bound[label_val, 1], t)
                        min_bound[label_val, 2] = min(min_bound[label_val, 2], z)
                        min_bound[label_val, 3] = min(min_bound[label_val, 3], y)
                        min_bound[label_val, 4] = min(min_bound[label_val, 4], x)

                        max_bound[label_val, 0] = max(max_bound[label_val, 0], s)
                        max_bound[label_val, 1] = max(max_bound[label_val, 1], t)
                        max_bound[label_val, 2] = max(max_bound[label_val, 2], z)
                        max_bound[label_val, 3] = max(max_bound[label_val, 3], y)
                        max_bound[label_val, 4] = max(max_bound[label_val, 4], x)

    return min_bound, max_bound


def calc_bounds4(cnp.ndarray[label_types, ndim=4] labels, Py_ssize_t components_num):
    """
    Calculate the bounds of the specified labels in a 4-dimensional array.

    Parameters
    ----------
    labels: numpy.ndarray[label_types, ndim=4]
        The 4-dimensional array containing the labels.
    components_num: Py_ssize_t
        The number of components to calculate the bounds for.

    Returns
    -------
    bounds: tuple[numpy.ndarray[cnp.int16_t, ndim=2], numpy.ndarray[cnp.int16_t, ndim=2]]
        The bounds of each component.
    """
    cdef Py_ssize_t x, y, z, t
    cdef Py_ssize_t x_max = labels.shape[3]
    cdef Py_ssize_t y_max = labels.shape[2]
    cdef Py_ssize_t z_max = labels.shape[1]
    cdef Py_ssize_t t_max = labels.shape[0]
    cdef label_types label_val
    cdef cnp.ndarray[cnp.int16_t, ndim=2] min_bound = np.full((components_num + 1, 4), max(x_max, y_max, z_max, t_max) + 5, dtype=np.int16)
    cdef cnp.ndarray[cnp.int16_t, ndim=2] max_bound = np.full((components_num + 1, 4), -1, dtype=np.int16)

    for t in range(0, t_max):
        for z in range(0, z_max):
            for y in range(0,y_max):
                for x in range(0,x_max):
                    label_val = labels[t, z, y, x]
                    if label_val == 0:
                        continue
                    min_bound[label_val, 0] = min(min_bound[label_val, 0], t)
                    min_bound[label_val, 1] = min(min_bound[label_val, 1], z)
                    min_bound[label_val, 2] = min(min_bound[label_val, 2], y)
                    min_bound[label_val, 3] = min(min_bound[label_val, 3], x)

                    max_bound[label_val, 0] = max(max_bound[label_val, 0], t)
                    max_bound[label_val, 1] = max(max_bound[label_val, 1], z)
                    max_bound[label_val, 2] = max(max_bound[label_val, 2], y)
                    max_bound[label_val, 3] = max(max_bound[label_val, 3], x)
    return min_bound, max_bound

def calc_bounds3(cnp.ndarray[label_types, ndim=3] labels, Py_ssize_t components_num):
    """
    Calculate the bounds of the specified labels in a 3-dimensional array.

    Parameters
    ----------
    labels: numpy.ndarray[label_types, ndim=3]
        The 3-dimensional array containing the labels.
    components_num: Py_ssize_t
        The number of components to calculate the bounds for.

    Returns
    -------
    bounds: tuple[numpy.ndarray[cnp.int16_t, ndim=2], numpy.ndarray[cnp.int16_t, ndim=2]]
        The bounds of each component.
    """
    cdef Py_ssize_t x, y, z
    cdef Py_ssize_t x_max = labels.shape[2]
    cdef Py_ssize_t y_max = labels.shape[1]
    cdef Py_ssize_t z_max = labels.shape[0]
    cdef label_types label_val
    cdef cnp.ndarray[cnp.int16_t, ndim=2] min_bound = np.full((components_num + 1, 3), max(x_max, y_max, z_max) + 5, dtype=np.int16)
    cdef cnp.ndarray[cnp.int16_t, ndim=2] max_bound = np.full((components_num + 1, 3), -1, dtype=np.int16)


    for z in range(0, z_max):
        for y in range(0,y_max):
            for x in range(0,x_max):
                label_val = labels[z, y, x]
                if label_val == 0:
                    continue
                min_bound[label_val, 0] = min(min_bound[label_val, 0], z)
                min_bound[label_val, 1] = min(min_bound[label_val, 1], y)
                min_bound[label_val, 2] = min(min_bound[label_val, 2], x)

                max_bound[label_val, 0] = max(max_bound[label_val, 0], z)
                max_bound[label_val, 1] = max(max_bound[label_val, 1], y)
                max_bound[label_val, 2] = max(max_bound[label_val, 2], x)
    return min_bound, max_bound


def calc_bounds2(cnp.ndarray[label_types, ndim=2] labels, Py_ssize_t components_num):
    """
    Calculate the bounds of the specified labels in a 2-dimensional array.

    Parameters
    ----------
    labels: numpy.ndarray[label_types, ndim=2]
        The 2-dimensional array containing the labels.
    components_num: Py_ssize_t
        The number of components to calculate the bounds for.

    Returns
    -------
    bounds: tuple[numpy.ndarray[cnp.int16_t, ndim=2], numpy.ndarray[cnp.int16_t, ndim=2]]
        The bounds of each component.
    """
    cdef Py_ssize_t x, y
    cdef Py_ssize_t x_max = labels.shape[1]
    cdef Py_ssize_t y_max = labels.shape[0]
    cdef label_types label_val
    cdef cnp.ndarray[cnp.int16_t, ndim=2] min_bound = np.full((components_num + 1, 2), max(x_max, y_max) + 5, dtype=np.int16)
    cdef cnp.ndarray[cnp.int16_t, ndim=2] max_bound = np.full((components_num + 1, 2), -1, dtype=np.int16)


    for y in range(0,y_max):
        for x in range(0,x_max):
            label_val = labels[y, x]
            if label_val == 0:
                continue
            min_bound[label_val, 0] = min(min_bound[label_val, 0], y)
            min_bound[label_val, 1] = min(min_bound[label_val, 1], x)

            max_bound[label_val, 0] = max(max_bound[label_val, 0], y)
            max_bound[label_val, 1] = max(max_bound[label_val, 1], x)

    return min_bound, max_bound



import typing
from typing import Any

import numpy as np

from PartSegCore_compiled_backend.sprawl_utils.euclidean_cython import calculate_euclidean as calculate_euclidean
from PartSegCore_compiled_backend.sprawl_utils.fuzzy_distance import MuType as MuType
from PartSegCore_compiled_backend.sprawl_utils.fuzzy_distance import calculate_mu_array as calculate_mu_array
from PartSegCore_compiled_backend.sprawl_utils.fuzzy_distance import fuzzy_distance as fuzzy_distance
from PartSegCore_compiled_backend.sprawl_utils.path_sprawl_cython import calculate_maximum as calculate_maximum
from PartSegCore_compiled_backend.sprawl_utils.path_sprawl_cython import calculate_minimum as calculate_minimum
from PartSegCore_compiled_backend.sprawl_utils.sprawl_utils import get_closest_component as get_closest_component
from PartSegCore_compiled_backend.sprawl_utils.sprawl_utils import get_maximum_component as get_maximum_component
from PartSegCore_compiled_backend.sprawl_utils.sprawl_utils import get_minimum_component as get_minimum_component

def path_maximum_sprawl(
    data_f: np.ndarray,
    components: np.ndarray,
    components_count: int,
    neighbourhood: np.ndarray,
    distance_cache: Any | None = ...,
    data_cache: Any | None = ...,
) -> np.ndarray: ...
def path_minimum_sprawl(
    data_f, components, components_count, neighbourhood, distance_cache: Any | None = ..., data_cache: Any | None = ...
) -> np.ndarray: ...
def euclidean_sprawl(
    data_m: np.ndarray,
    components: np.ndarray,
    components_count: int,
    neigh_arr,
    dist_arr,
    distance_cache: Any | None = ...,
    data_cache: Any | None = ...,
) -> np.ndarray: ...
def fdt_sprawl(
    data_m: np.ndarray,
    components: np.ndarray,
    components_count: int,
    neigh_arr,
    dist_arr,
    lower_bound,
    upper_bound,
    distance_cache: Any | None = ...,
    data_cache: Any | None = ...,
) -> np.ndarray: ...
def sprawl_component(
    data_m: np.ndarray,
    components: np.ndarray,
    component_number: int,
    neigh_arr: np.ndarray,
    dist_arr: np.ndarray,
    calculate_operator: typing.Callable,
) -> np.ndarray: ...
def distance_sprawl(
    calculate_operator,
    data_m: np.ndarray,
    components: np.ndarray,
    components_count: int,
    neigh_arr,
    dist_array,
    distance_cache: Any | None = ...,
    data_cache: Any | None = ...,
    parallel: bool = ...,
) -> np.ndarray: ...
def reverse_permutation(perm: list[int]) -> list[int]: ...
def relabel_with_perm(labeling: list, perm: list) -> list: ...
def verify_cohesion(elements: list[int], graph: list[list[int]]) -> bool: ...
def relabel_array(data: np.ndarray, perm: list[int]) -> np.ndarray: ...



# distutils: language = c++
# cython: wraparound=False, nonecheck=False, cdivision=True, embedsignature=True, unraisable_tracebacks=True
# cython: profile=True
# cython: language_level=3

import numpy as np
cimport numpy as np

ctypedef fused image_types:
    np.float32_t
    np.float64_t
    np.uint8_t
    np.uint16_t
    np.uint32_t

ctypedef fused component_types:
    np.uint32_t
    np.uint16_t
    np.uint8_t


def get_maximum_component(components, data_mask, paths, components_translation, num_of_components=None):
    if num_of_components is None:
        num_of_components = len(paths)
    new_paths = paths.reshape((paths.size//components.size, components.size))
    _get_maximum_component(components.ravel(), data_mask.ravel(), new_paths, num_of_components, components_translation)
    return components


def _get_maximum_component(np.ndarray[component_types] components, np.ndarray[np.uint8_t] data_mask,
                           np.ndarray[image_types, ndim=2] paths, int num_of_components,
                           np.ndarray[np.uint32_t] components_translation):
    cdef Py_ssize_t x, y, size, component_index
    cdef image_types max_val, val
    cdef char flag
    size = components.size
    for x in range(size):
        flag = False
        if components[x] == 0 and data_mask[x]:
            max_val = paths[0, x]
            component_index = components[0]
            flag = True
            for y in range(1, num_of_components+1):
                val = paths[y, x]
                if val > max_val:
                    max_val = val
                    component_index = components_translation[y]
                    flag = True
                elif val == max_val:
                    flag = False
            if flag:
                components[x] = component_index
                paths[0, x] = max_val
    return components

def get_minimum_component(components, data_mask, paths, components_translation, num_of_components=None):
    if num_of_components is None:
        num_of_components = len(paths)
    new_paths = paths.reshape((paths.size//components.size, components.size))
    _get_minimum_component(components.ravel(), data_mask.ravel(), new_paths, num_of_components, components_translation)
    return components


def _get_minimum_component(np.ndarray[component_types] components, np.ndarray[np.uint8_t] data_mask,
                           np.ndarray[image_types, ndim=2] paths, int num_of_components,
                           np.ndarray[np.uint32_t] components_translation):
    cdef Py_ssize_t x, y, size, component_index
    cdef image_types min_val, val
    cdef char flag
    size = components.size
    for x in range(size):
        flag = False
        if components[x] == 0 and data_mask[x]:
            min_val = paths[0, x]
            component_index = components[0]
            flag = True
            for y in range(1, num_of_components+1):
                val = paths[y, x]
                if val < min_val:
                    min_val = val
                    component_index = components_translation[y]
                    flag = True
                elif val == min_val:
                    flag = False
            if flag:
                components[x] = component_index
                paths[0, x] = min_val
    return components


def get_closest_component(components, data_mask, distances, components_translation, num_of_components=None):
    if num_of_components is None:
        num_of_components = len(distances)
    new_distances = distances.reshape((distances.size // components.size, components.size))
    _get_closest_component(components.ravel(), data_mask.ravel(), new_distances, num_of_components,
                           components_translation)
    return components


def _get_closest_component(np.ndarray[component_types] components, np.ndarray[np.uint8_t] data_mask,
                           np.ndarray[image_types, ndim=2] paths, int num_of_components,
                           np.ndarray[np.uint32_t] components_translation):
    cdef Py_ssize_t x, y, size, component_index, count1=0, count2=0, count3=0
    cdef image_types min_val, val
    cdef char flag
    size = components.size
    for x in range(size):
        flag = False
        if data_mask[x]:
            min_val = paths[0, x]
            component_index = components[x]
            flag = True
            for y in range(1, num_of_components+1):
                count3 += 1
                val = paths[y, x]
                if val < min_val:
                    count1 += 1
                    min_val = val
                    component_index = components_translation[y]
                    flag = True
                elif val == min_val:
                    flag = True
                    component_index = 0
            if flag:
                count2 += 1
                components[x] = component_index
                paths[0, x] = min_val
    return components, paths



# distutils: define_macros=CYTHON_TRACE_NOGIL=1
# distutils: language = c++
# cython: language_level=3, boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, embedsignature=True

from collections.abc import Sequence

import numpy as np
import cython

cimport numpy as cnp

from libcpp cimport bool
from libcpp.unordered_set cimport unordered_set
from libcpp.utility cimport pair
from libcpp.vector cimport vector


cdef extern from "triangulation/point.hpp" namespace "partsegcore::point":
    cdef cppclass Point:
        float x
        float y
        Point()
        Point(float x, float y)
        bool operator==(const Point& other) const
        bool operator!=(const Point& other) const

    cdef cppclass Vector:
        float x;
        float y;

    cdef cppclass Segment:
        Point bottom
        Point top
        Segment()
        Segment(Point bottom, Point top)


cdef extern from "triangulation/intersection.hpp" namespace "partsegcore::intersection":
    cdef cppclass Event:
        float x
        float y
        int index
        bool is_top
        Event()
        Event(float x, float y, int index, bool is_top)

    cdef cppclass OrderedPair:
        int first
        int second
        OrderedPair()
        OrderedPair(int first, int second)

    bool _on_segment(const Point& p, const Point& q, const Point& r)
    int _orientation(const Point& p, const Point& q, const Point& r)
    bool _do_intersect(const Segment& s1, const Segment& s2)
    unordered_set[OrderedPair] _find_intersections(const vector[Segment]& segments)
    vector[Point] _find_intersection(const Segment& s1, const Segment& s2)


cdef extern from "triangulation/triangulate.hpp" namespace "partsegcore::triangulation":

    cdef cppclass Triangle:
        size_t x
        size_t y
        size_t z
        Triangle()
        Triangle(int x, int y, int z)

    cdef cppclass MonotonePolygon:
        Point top
        Point bottom
        vector[Point] left
        vector[Point] right
        MonotonePolygon()
        MonotonePolygon(Point top, Point bottom, vector[Point] left, vector[Point] right)


    cdef cppclass PointTriangle:
        Point p1
        Point p2
        Point p3

    cdef cppclass PathTriangulation:
        vector[Triangle] triangles
        vector[Point] centers
        vector[Vector] offsets


    bool _is_convex(const vector[Point]& polygon)
    vector[Triangle] _triangle_convex_polygon(const vector[Point]& polygon)
    bool left_to_right(const Segment& s1, const Segment& s2)
    vector[Point] find_intersection_points(const vector[Point]& segments)
    vector[PointTriangle] triangulate_monotone_polygon(const MonotonePolygon& polygon)
    pair[vector[Triangle], vector[Point]] triangulate_polygon_face(const vector[Point]& polygon) except + nogil
    pair[vector[Triangle], vector[Point]] triangulate_polygon_face(const vector[vector[Point]]& polygon_list) except + nogil
    PathTriangulation triangulate_path_edge(const vector[Point]& path, bool closed, float limit, bool bevel) except + nogil
    vector[vector[Point]] split_polygon_on_repeated_edges(const vector[Point]& polygon) except + nogil


ctypedef fused float_types:
    cnp.float32_t
    cnp.float64_t


def on_segment(p: Sequence[float], q: Sequence[float], r: Sequence[float]) -> bool:
    """ Check if point q is on segment pr

    Parameters
    ----------
    p: sequence of 2 floats
        beginning of segment
    q: sequence of 2 floats
        point to check
    r: sequence of 2 floats
        end of segment

    Returns
    -------
    bool:
        True if q is on segment pr
    """
    return _on_segment(
        Point(p[0], p[1]),
        Point(q[0], q[1]),
        Point(r[0], r[1])
        )

def orientation(p: Sequence[float], q: Sequence[float], r: Sequence[float]) -> int:
    """ Check orientation of 3 points

    Parameters
    ----------
    p: sequence of 2 floats
        first point
    q: sequence of 2 floats
        second point
    r: sequence of 2 floats
        third point

    Returns
    -------
    int:
        0 if points are collinear, 1 if clockwise, 2 if counterclockwise
    """
    return _orientation(
        Point(p[0], p[1]),
        Point(q[0], q[1]),
        Point(r[0], r[1])
        )


def do_intersect(s1: Sequence[Sequence[float]], s2: Sequence[Sequence[float]]) -> bool:
    """ Check if two segments intersect

    Parameters
    ----------
    s1: sequence of 2 sequences of 2 floats
        first segment
    s2: sequence of 2 sequences of 2 floats
        second segment

    Returns
    -------
    bool:
        True if segments intersect
    """
    return _do_intersect(
        Segment(Point(s1[0][0], s1[0][1]), Point(s1[1][0], s1[1][1])),
        Segment(Point(s2[0][0], s2[0][1]), Point(s2[1][0], s2[1][1]))
        )


def find_intersections(segments: Sequence[Sequence[Sequence[float]]]) -> list[tuple[int, int]]:
    """ Find intersections between segments"""
    cdef vector[Segment] segments_vector
    cdef unordered_set[OrderedPair] intersections
    cdef OrderedPair p

    segments_vector.reserve(len(segments))
    for segment in segments:
        segments_vector.push_back(Segment(Point(segment[0][0], segment[0][1]), Point(segment[1][0], segment[1][1])))
    intersections = _find_intersections(segments_vector)

    return [(p.first, p.second) for p  in intersections]


def find_intersection_point(s1: Sequence[Sequence[float]], s2: Sequence[Sequence[float]]) -> list[tuple[float, float]]:
    """ Find intersection between two segments

    Parameters
    ----------
    s1: sequence of 2 sequences of 2 floats
        first segment
    s2: sequence of 2 sequences of 2 floats
        second segment

    Returns
    -------
    sequence of 2 floats:
        intersection point
    """
    cdef vector[Point] p_li = _find_intersection(
        Segment(Point(s1[0][0], s1[0][1]), Point(s1[1][0], s1[1][1])),
        Segment(Point(s2[0][0], s2[0][1]), Point(s2[1][0], s2[1][1]))
        )
    return [(p.x, p.y) for p in p_li]


def is_convex(polygon: Sequence[Sequence[float]]) -> bool:
    """ Check if polygon is convex"""
    cdef vector[Point] polygon_vector
    cdef pair[bool, vector[int]] result

    polygon_vector.reserve(len(polygon))
    for point in polygon:
        polygon_vector.push_back(Point(point[0], point[1]))

    return _is_convex(polygon_vector)

def triangle_convex_polygon(polygon: Sequence[Sequence[float]])  -> list[tuple[int, int, int]]:
    cdef vector[Point] polygon_vector
    cdef vector[Triangle] result
    cdef Point p1, p2

    polygon_vector.reserve(len(polygon))
    polygon_vector.push_back(Point(polygon[0][0], polygon[0][1]))
    for point in polygon[1:]:
        p1 = polygon_vector[polygon_vector.size() - 1]
        p2 = Point(point[0], point[1])
        if p1 != p2:
            # prevent from adding polygon edge of width 0
            polygon_vector.push_back(p2)

    result = _triangle_convex_polygon(polygon_vector)
    return [(triangle.x, triangle.y, triangle.z) for triangle in result]



def triangulate_polygon_py(polygon: Sequence[Sequence[float]]) -> tuple[list[tuple[int, int, int]], list[tuple[float, float]]]:
    """ Triangulate polygon"""
    cdef vector[Point] polygon_vector
    cdef Point p1, p2
    cdef pair[vector[Triangle], vector[Point]] result

    polygon_vector.reserve(len(polygon))
    polygon_vector.push_back(Point(polygon[0][0], polygon[0][1]))
    for point in polygon[1:]:
        p1 = polygon_vector[polygon_vector.size() - 1]
        p2 = Point(point[0], point[1])
        if p1 != p2:
            # prevent from adding polygon edge of width 0
            polygon_vector.push_back(p2)

    result = triangulate_polygon_face(polygon_vector)
    return [(triangle.x, triangle.y, triangle.z) for triangle in result.first], [(point.x, point.y) for point in result.second]


def triangulate_polygon_numpy(cnp.ndarray[float_types, ndim=2] polygon: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """ Triangulate polygon"""
    cdef vector[Point] polygon_vector
    cdef Point p1, p2
    cdef pair[vector[Triangle], vector[Point]] result

    polygon_vector.reserve(polygon.shape[0])
    polygon_vector.push_back(Point(polygon[0, 0], polygon[0, 1]))
    for point in polygon[1:]:
        p1 = polygon_vector[polygon_vector.size() - 1]
        p2 = Point(point[0], point[1])
        if p1 != p2:
            # prevent from adding polygon edge of width 0
            polygon_vector.push_back(p2)

    result = triangulate_polygon_face(polygon_vector)
    return (
        np.array([(triangle.x, triangle.y, triangle.z) for triangle in result.first], dtype=np.uintp),
        np.array([(point.x, point.y) for point in result.second], dtype=np.float32)
    )


def triangulate_polygon_numpy_li(list polygon_li: list[np.ndarray]) -> tuple[np.ndarray, np.ndarray]:
    """ Triangulate polygon"""
    cdef vector[Point] polygon_vector
    cdef vector[vector[Point]] polygon_vector_list
    cdef Point p1, p2
    cdef pair[vector[Triangle], vector[Point]] result
    cdef size_t i;
    cdef cnp.ndarray[cnp.float32_t, ndim=2] polygon32
    cdef cnp.ndarray[cnp.float64_t, ndim=2] polygon64

    polygon_vector_list.reserve(len(polygon_li))
    for polygon in polygon_li:
        polygon_vector.clear()
        polygon_vector.reserve(polygon.shape[0])
        polygon_vector.push_back(Point(polygon[0, 0], polygon[0, 1]))

        if polygon.dtype == np.float32:
            polygon32 = polygon
            for i in range(1, polygon.shape[0]):
                p1 = polygon_vector.back()
                p2 = Point(polygon32[i, 0], polygon32[i, 1])
                if p1 != p2:
                    # prevent from adding polygon edge of width 0
                    polygon_vector.push_back(p2)
        else:
            polygon64 = polygon
            for i in range(1, polygon.shape[0]):
                p1 = polygon_vector.back()
                p2 = Point(polygon64[i, 0], polygon64[i, 1])
                if p1 != p2:
                    # prevent from adding polygon edge of width 0
                    polygon_vector.push_back(p2)

        if polygon_vector.size() > 1 and polygon_vector.front() == polygon_vector.back():
            polygon_vector.pop_back()
        polygon_vector_list.push_back(polygon_vector)

    result = triangulate_polygon_face(polygon_vector_list)
    return (
        np.array([(triangle.x, triangle.y, triangle.z) for triangle in result.first], dtype=np.uintp),
        np.array([(point.x, point.y) for point in result.second], dtype=np.float32)
    )

def segment_left_to_right_comparator(s1: Sequence[Sequence[float]], s2: Sequence[Sequence[float]]) -> bool:
    """ Compare segments by bottom point"""
    return left_to_right(
        Segment(Point(s1[0][0], s1[0][1]), Point(s1[1][0], s1[1][1])),
        Segment(Point(s2[0][0], s2[0][1]), Point(s2[1][0], s2[1][1]))
        )


def find_intersection_points_py(polygon: Sequence[Sequence[float]]) -> Sequence[Sequence[float]]:
    """ Find intersection points in polygon"""
    cdef vector[Point] polygon_vector
    cdef vector[Point] result

    polygon_vector.reserve(len(polygon))
    for point in polygon:
        polygon_vector.push_back(Point(point[0], point[1]))

    result = find_intersection_points(polygon_vector)
    return [(point.x, point.y) for point in result]


def triangulate_monotone_polygon_py(top: Sequence[float], bottom: Sequence[float], left: Sequence[Sequence[float]], right: Sequence[Sequence[float]]) -> Sequence[Sequence[Sequence[float]]]:
    """ Triangulate monotone polygon"""
    cdef vector[Point] left_vector, right_vector
    cdef vector[PointTriangle] result
    cdef MonotonePolygon mono_polygon

    left_vector.reserve(len(left))
    for point in left:
        left_vector.push_back(Point(point[0], point[1]))

    right_vector.reserve(len(right))
    for point in right:
        right_vector.push_back(Point(point[0], point[1]))

    mono_polygon = MonotonePolygon(Point(top[0], top[1]), Point(bottom[0], bottom[1]), left_vector, right_vector)
    result = triangulate_monotone_polygon(mono_polygon)
    return [
        [(triangle.p1.x, triangle.p1.y), (triangle.p2.x, triangle.p2.y), (triangle.p3.x, triangle.p3.y)]
        for triangle in result
    ]


def triangulate_path_edge_py(path: Sequence[Sequence[float]], closed: bool=False, limit: float=3.0, bevel: bool=False) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """ Triangulate path"""
    cdef vector[Point] path_vector
    cdef PathTriangulation result
    cdef Point p1, p2
    cdef cnp.ndarray[cnp.uint32_t, ndim=2] triangles

    path_vector.reserve(len(path))
    for point in path:
        path_vector.push_back(Point(point[0], point[1]))
    with cython.nogil:
        result = triangulate_path_edge(path_vector, closed, limit, bevel)

    if result.triangles.size() == 0:
        triangles = np.zeros((0, 3), dtype=np.uint32)
    else:
        triangles = np.array([(triangle.x, triangle.y, triangle.z) for triangle in result.triangles], dtype=np.uint32)

    return (
        np.array([(point.x, point.y) for point in result.centers], dtype=np.float32),
        np.array([(offset.x, offset.y) for offset in result.offsets], dtype=np.float32),
        triangles,
    )

def triangulate_path_edge_numpy(cnp.ndarray[cnp.float32_t, ndim=2] path, bool closed=False, float limit=3.0, bool bevel=False) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """ Triangulate path"""
    cdef vector[Point] path_vector
    cdef PathTriangulation result
    cdef Point p1, p2
    cdef cnp.ndarray[cnp.uint32_t, ndim=2] triangles
    cdef cnp.ndarray[cnp.float32_t, ndim=2] offsets, centers
    cdef size_t i, len_path

    len_path = path.shape[0]

    path_vector.reserve(len_path)
    for i in range(len_path):
        path_vector.push_back(Point(path[i, 0], path[i, 1]))
    with cython.nogil:
        result = triangulate_path_edge(path_vector, closed, limit, bevel)

    triangles = np.empty((result.triangles.size(), 3), dtype=np.uint32)
    centers = np.empty((result.centers.size(), 2), dtype=np.float32)
    offsets = np.empty((result.offsets.size(), 2), dtype=np.float32)

    for i in range(result.triangles.size()):
        triangles[i, 0] = result.triangles[i].x
        triangles[i, 1] = result.triangles[i].y
        triangles[i, 2] = result.triangles[i].z

    for i in range(result.centers.size()):
        centers[i, 0] = result.centers[i].x
        centers[i, 1] = result.centers[i].y

    for i in range(result.offsets.size()):
        offsets[i, 0] = result.offsets[i].x
        offsets[i, 1] = result.offsets[i].y

    return (
        centers,
        offsets,
        triangles,
    )


def triangulate_polygon_with_edge_numpy_li(polygon_li: list[np.ndarray], split_edges: bool=False) -> tuple[tuple[np.ndarray, np.ndarray], tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """ Triangulate polygon"""
    cdef vector[Point] polygon_vector
    cdef vector[vector[Point]] polygon_vector_list, edge_split_list
    cdef Point p1, p2
    cdef pair[vector[Triangle], vector[Point]] triangulation_result
    cdef vector[PathTriangulation] edge_result
    cdef cnp.ndarray[cnp.uint32_t, ndim=2] triangles, edge_triangles
    cdef size_t triangle_count = 0

    cdef cnp.ndarray[cnp.float32_t, ndim=2] points, edge_offsets, edges_centers, polygon
    cdef size_t i, j, len_path, edge_triangle_count, edge_center_count, edge_triangle_index, edge_center_index

    polygon_vector_list.reserve(len(polygon_li))
    for i in range(len(polygon_li)):
        polygon = polygon_li[i]

        polygon_vector.reserve(polygon.shape[0])
        polygon_vector.push_back(Point(polygon[0, 0], polygon[0, 1]))

        for j in range(1, polygon.shape[0]):
            p1 = polygon_vector.back()
            p2 = Point(polygon[j, 0], polygon[j, 1])
            if p1 != p2:
                # prevent from adding polygon edge of width 0
                polygon_vector.push_back(p2)
        if polygon_vector.size() > 1 and polygon_vector.front() == polygon_vector.back():
            polygon_vector.pop_back()
        polygon_vector_list.push_back(polygon_vector)
        if split_edges:
            with cython.nogil:
                edge_split_list = split_polygon_on_repeated_edges(polygon_vector)
                for edge_li in edge_split_list:
                    edge_result.push_back(triangulate_path_edge(edge_li, True, 3.0, False))
        else:
            with cython.nogil:
                edge_result.push_back(triangulate_path_edge(polygon_vector, True, 3.0, False))

    with cython.nogil:
        triangulation_result = triangulate_polygon_face(polygon_vector_list)


    triangles = np.empty((triangulation_result.first.size(), 3), dtype=np.uint32)
    for i in range(triangulation_result.first.size()):
        triangles[i, 0] = triangulation_result.first[i].x
        triangles[i, 1] = triangulation_result.first[i].y
        triangles[i, 2] = triangulation_result.first[i].z

    points = np.empty((triangulation_result.second.size(), 2), dtype=np.float32)
    for i in range(triangulation_result.second.size()):
        points[i, 0] = triangulation_result.second[i].x
        points[i, 1] = triangulation_result.second[i].y

    edge_triangle_count = 0
    edge_center_count = 0
    for i in range(edge_result.size()):
        edge_triangle_count += edge_result[i].triangles.size()
        edge_center_count += edge_result[i].centers.size()

    edge_triangles = np.empty((edge_triangle_count, 3), dtype=np.uint32)
    edge_offsets = np.empty((edge_center_count, 2), dtype=np.float32)
    edges_centers = np.empty((edge_center_count, 2), dtype=np.float32)

    edge_triangle_index = 0
    edge_center_index = 0
    for i in range(edge_result.size()):
        for j in range(edge_result[i].triangles.size()):
            edge_triangles[edge_triangle_index, 0] = edge_result[i].triangles[j].x + triangle_count
            edge_triangles[edge_triangle_index, 1] = edge_result[i].triangles[j].y + triangle_count
            edge_triangles[edge_triangle_index, 2] = edge_result[i].triangles[j].z + triangle_count
            edge_triangle_index += 1
        triangle_count += edge_result[i].centers.size()

        for j in range(edge_result[i].centers.size()):
            edges_centers[edge_center_index, 0] = edge_result[i].centers[j].x
            edges_centers[edge_center_index, 1] = edge_result[i].centers[j].y
            edge_offsets[edge_center_index, 0] = edge_result[i].offsets[j].x
            edge_offsets[edge_center_index, 1] = edge_result[i].offsets[j].y
            edge_center_index += 1

    return ((
        triangles,
        points,
    ),
    (
        edges_centers,
        edge_offsets,
        edge_triangles,
    )
    )


def split_polygon_on_repeated_edges_py(polygon: Sequence[Sequence[float]]) -> list[list[tuple[float, float]]]:
    """ Split polygon on repeated edges"""
    cdef vector[Point] polygon_vector
    cdef vector[vector[Point]] result
    cdef Point p1, p2

    polygon_vector.reserve(len(polygon))
    for point in polygon:
        polygon_vector.push_back(Point(point[0], point[1]))

    result = split_polygon_on_repeated_edges(polygon_vector)
    return [
        [(point.x, point.y) for point in polygon_vector]
        for polygon_vector in result
    ]



# distutils: define_macros=CYTHON_TRACE_NOGIL=1
# cython: language_level=3, boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, embedsignature=True

import numpy as np

cimport numpy as np

from cython.parallel import prange

from libcpp cimport bool


cpdef enum:
    resolution = 1024
"""current accuracy of coloring"""

cdef extern from "<algorithm>" namespace "std" nogil:
    T max[T](const T& v1, const T& v2)
    T min[T](const T& v1, const T& v2)

DTYPE = np.uint8
ctypedef np.uint8_t DTYPE_t

ctypedef fused numpy_types:
    np.uint8_t
    np.uint16_t
    np.uint32_t
    np.int8_t
    np.int16_t
    np.int32_t
    np.int64_t
    np.float32_t
    np.float64_t

ctypedef fused label_types:
    np.uint8_t
    np.uint16_t
    np.uint32_t

cdef inline long scale_factor(numpy_types value, double min_val, double factor) nogil:
    return max[long](0, min[long](resolution - 1, <long>((value - min_val) * factor)))


def calculate_borders(np.ndarray[label_types, ndim=4] labels, int border_thick, bool per_layer=True):
    if labels.shape[1] < 3:
        per_layer = True
    if per_layer:
        return calculate_borders2d(labels, border_thick)
    cdef Py_ssize_t x, y, z, t
    cdef int R, Y, Z
    cdef Py_ssize_t x_max = labels.shape[3]
    cdef Py_ssize_t y_max = labels.shape[2]
    cdef Py_ssize_t z_max = labels.shape[1]
    cdef Py_ssize_t t_max = labels.shape[0]
    cdef Py_ssize_t circle_len
    cdef np.ndarray[np.int8_t, ndim=2] circle_shift
    cdef np.ndarray[label_types, ndim=4] local_labels
    cdef label_types label_val

    R = border_thick**2
    values = list(range(-border_thick, border_thick+1))
    circle_list = []
    for x in values:
        for y in values:
            for z in values:
                if (x**2 + y**2 + z **2 <= R):
                    circle_list.append((x,y,z))
    circle_shift = np.array(circle_list).astype(np.int8)
    circle_len = circle_shift.shape[0]
    local_labels = np.zeros((labels.shape[0], labels.shape[1] + 2*border_thick, labels.shape[2] + 2*border_thick, labels.shape[3] + 2*border_thick), dtype=labels.dtype)
    circle_shift = circle_shift + border_thick
    for t in range(0, t_max):
        for z in range(1, z_max-1):
            for y in range(1,y_max-1):
                for x in range(1,x_max-1):
                    label_val = labels[t, z, y, x]
                    if label_val and (labels[t, z, y+1, x] != label_val or labels[t, z, y-1, x] != label_val or
                                      labels[t, z, y, x+1] != label_val or labels[t, z, y, x-1] != label_val or
                                      labels[t, z+1, y, x] != label_val or labels[t, z-1, y, x] != label_val):
                        for circle_pos in range(circle_len):
                            local_labels[t, z + circle_shift[circle_pos, 0], y + circle_shift[circle_pos, 1], x + circle_shift[circle_pos, 2]] = label_val
    if border_thick > 0:
        local_labels = local_labels[:, border_thick:-border_thick, border_thick:-border_thick, border_thick:-border_thick]
    return local_labels


def calculate_borders2d(np.ndarray[label_types, ndim=4] labels, int border_thick):
    cdef Py_ssize_t x, y, z, t
    cdef int R, Y, Z
    cdef Py_ssize_t x_max = labels.shape[3]
    cdef Py_ssize_t y_max = labels.shape[2]
    cdef Py_ssize_t z_max = labels.shape[1]
    cdef Py_ssize_t t_max = labels.shape[0]
    cdef Py_ssize_t circle_len
    cdef np.ndarray[np.int8_t, ndim=2] circle_shift
    cdef np.ndarray[label_types, ndim=4] local_labels
    cdef label_types label_val

    R = border_thick**2
    values = list(range(-border_thick, border_thick+1))
    circle_list = []
    for x in values:
        for y in values:
            if (x**2 + y**2 <= R):
                circle_list.append((x,y))
    circle_shift = np.array(circle_list).astype(np.int8)
    circle_len = circle_shift.shape[0]
    local_labels = np.zeros((labels.shape[0], labels.shape[1], labels.shape[2] + 2*border_thick, labels.shape[3] + 2*border_thick), dtype=labels.dtype)
    circle_shift = circle_shift + border_thick
    for t in range(0, t_max):
        for z in range(0, z_max):
            for y in range(1,y_max-1):
                for x in range(1,x_max-1):
                    label_val = labels[t, z, y, x]
                    if label_val and (labels[t, z, y+1, x] != label_val or labels[t, z, y-1, x] != label_val or
                                      labels[t, z, y, x+1] != label_val or labels[t, z, y, x-1] != label_val):
                        for circle_pos in range(circle_len):
                            local_labels[t, z, y + circle_shift[circle_pos, 0], x + circle_shift[circle_pos, 1]] =\
                                label_val
    if border_thick > 0:
        local_labels = local_labels[:, :, border_thick:-border_thick, border_thick:-border_thick]
    return local_labels


def color_grayscale(np.ndarray[DTYPE_t, ndim=2] cmap, np.ndarray[numpy_types, ndim=2] image, double min_val,
                    double max_val, int single_channel=-1):
    """color image channel in respect to cmap array. Array should be in size (resolution, 3)"""
    # TODO maybe not use empty channels
    cdef Py_ssize_t x_max = image.shape[0]
    cdef Py_ssize_t y_max = image.shape[1]
    cdef int val, x, y
    cdef double factor =  1/ ((max_val - min_val) / (resolution - 1))
    cdef np.ndarray[DTYPE_t, ndim=3] result_array = np.zeros((x_max, y_max, 3), dtype=DTYPE)
    with nogil:
        if 0 <= single_channel <= 2:
            for x in prange(x_max):
                for y in range(y_max):
                    val = scale_factor(image[x, y], min_val, factor)
                    result_array[x, y, single_channel] = cmap[val, single_channel]
        else:
            for x in prange(x_max):
                for y in range(y_max):
                    val = scale_factor(image[x, y], min_val, factor)
                    result_array[x, y, 0] = cmap[val, 0]
                    result_array[x, y, 1] = cmap[val, 1]
                    result_array[x, y, 2] = cmap[val, 2]

    return result_array

def add_labels(np.ndarray[DTYPE_t, ndim=3] image, np.ndarray[label_types, ndim=2] labels, float overlay, int only_border,
              int border_thick, np.ndarray[np.uint8_t, ndim=1] use_labels, np.ndarray[np.uint8_t, ndim=2] label_colors):
    """
    Add label to given RGB image. Ift modify original image.
    :param image: RGB image on background
    :param labels: labels, different label is different num
    :param overlay: from 0 to 1
    :param only_border: Draw only border of image
    :param border_thick: if draw only border then thick of this border
    :param use_labels: Which labels should be drawer
    :return: changed image
    """
    if labels is None:
        return image
    cdef Py_ssize_t x_max = image.shape[0]
    cdef Py_ssize_t y_max = image.shape[1]
    cdef Py_ssize_t x, y
    cdef Py_ssize_t comp_num = use_labels.size
    cdef int R, Y
    cdef Py_ssize_t circle_len, circle_pos, labels_colors_num
    cdef np.ndarray[np.int8_t, ndim=2] circle_shift
    cdef label_types label_val

    cdef np.ndarray[label_types, ndim=2] local_labels

    # cdef np.uint8_t [:,:] label_colors = label_colors_global
    labels_colors_num = label_colors.shape[0]
    # prevent from usage background
    use_labels[0] = 0

    if only_border:
        circle_list = []
        if border_thick > 0:
            R = border_thick
            for x in range(-R,R+1):
                Y = int((R*R-x*x)**0.5) # bound for y given x
                for y in range(-Y,Y+1):
                    circle_list.append([x,y])
            circle_shift = np.array(circle_list).astype(np.int8)
        else:
            circle_shift = np.array([[0,0]]).astype(np.int8)
        circle_shift += border_thick
        circle_len = circle_shift.shape[0]
        local_labels = np.zeros((labels.shape[0] + 2*border_thick, labels.shape[1] + 2*border_thick), dtype=labels.dtype)
        for x in range(1,x_max-1):
            for y in range(1,y_max-1):
                label_val =labels[x,y]
                if use_labels[label_val] and (labels[x+1,y] != label_val or labels[x-1, y] != label_val or labels[x, y+1] != label_val or labels[x, y-1] != label_val):
                    for circle_pos in range(circle_len):
                        local_labels[x + circle_shift[circle_pos, 0], y + circle_shift[circle_pos, 1]] = label_val

        if border_thick > 0:
            local_labels = local_labels[border_thick:-border_thick, border_thick:-border_thick]
    else:
        local_labels = np.copy(labels)
        for i in range(1, comp_num):
            if not use_labels[i]:
                local_labels[local_labels == i] = 0
    cdef float part1, part2
    cdef label_types col_num
    for x in prange(x_max, nogil=True):
        for y in range(y_max):
            if local_labels[x,y] > 0:
                col_num = (local_labels[x,y] -1) % labels_colors_num
                image[x, y, 0] = <DTYPE_t> (image[x, y, 0] * (1-overlay) + label_colors[col_num, 0] * overlay)
                image[x, y, 1] = <DTYPE_t> (image[x, y, 1] * (1-overlay) + label_colors[col_num, 1] * overlay)
                image[x, y, 2] = <DTYPE_t> (image[x, y, 2] * (1-overlay) + label_colors[col_num, 2] * overlay)
    # local_labels = labels_color_map[labels]
    return image



# distutils: define_macros=CYTHON_TRACE_NOGIL=1
# distutils: language = c++
# cython: language_level=3, boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, embedsignature=True

import numpy as np

cimport numpy as cnp

from cython.operator import dereference
from cython.parallel import prange
from libcpp.unordered_map cimport unordered_map

ctypedef fused label_types:
    cnp.uint8_t
    cnp.int8_t
    cnp.uint16_t
    cnp.int16_t
    cnp.uint32_t
    cnp.int32_t
    cnp.uint64_t
    cnp.int64_t

ctypedef fused out_types_mod:
    cnp.uint8_t
    cnp.uint16_t

ctypedef fused out_types:
    out_types_mod
    cnp.float32_t

def _zero_preserving_modulo_seq(label_types[:] labels,  int modulo, int to_zero, out_types_mod[:] out):
    cdef Py_ssize_t i
    cdef Py_ssize_t n = labels.shape[0]
    for i in range(n):
        if labels[i] == to_zero:
            out[i] = 0
        else:
            out[i] = ((modulo + ((labels[i] - 1) % modulo)) % modulo) + 1

def _zero_preserving_modulo_par(label_types[:] labels, int modulo, int to_zero, out_types_mod[:] out):
    cdef Py_ssize_t i
    cdef Py_ssize_t n = labels.shape[0]
    for i in prange(n, nogil=True):
        if labels[i] == to_zero:
            out[i] = 0
        else:
            out[i] = ((modulo + ((labels[i] - 1) % modulo)) % modulo) + 1


def _map_array_par(label_types[:] labels, dict dkt, out_types[:] out, out_types def_val=0):
    # build the map from the input and output vectors
    cdef Py_ssize_t i, n_map, n_array
    cdef unordered_map[label_types, out_types].iterator it
    cdef unordered_map[label_types, out_types] lut
    for key, val in dkt.items():
        if key is not None:
            lut[<label_types>key] = <out_types> val
        else:
            def_val = <out_types> val
    # apply the map to the array
    n_array = labels.shape[0]
    for i in prange(n_array, nogil=True): #
        it = lut.find(labels[i])
        if it != lut.end():
            out[i] = dereference(it).second
        else:
            out[i] = def_val


def _map_array_seq(label_types[:] labels, dict dkt, out_types[:] out, out_types def_val=0):
    # build the map from the input and output vectors
    cdef Py_ssize_t i, n_map, n_array
    cdef unordered_map[label_types, out_types].iterator it
    cdef unordered_map[label_types, out_types] lut
    for key, val in dkt.items():
        if key is not None:
            lut[<label_types>key] = <out_types> val
        else:
            def_val = <out_types> val
    # apply the map to the array
    n_array = labels.shape[0]
    for i in range(n_array): #
        it = lut.find(labels[i])
        if it != lut.end():
            out[i] = dereference(it).second
        else:
            out[i] = def_val

# cdef class MapWrapper:
#     cdef unordered_map[label_types, out_types] lut



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True
# cython: language_level=3

cimport numpy as cnp

from libcpp cimport bool


ctypedef cnp.int16_t Size

cdef struct Point:
    Size x
    Size y
    Size z

cdef extern from 'my_queue.h':
    cdef cppclass my_queue[T]:
        my_queue() except +
        T front()
        void pop()
        void push(T& v)
        bool empty()
        size_t get_size()

cdef extern from "global_consts.h":
    const signed char neighbourhood[26][3]
    const char neigh_level[]
    const float distance[]

ctypedef fused component_types:
    cnp.uint32_t
    cnp.uint16_t
    cnp.uint8_t



import numpy as np
import pytest

from PartSegCore_compiled_backend.calc_bounds import calc_bounds


@pytest.mark.parametrize('ndim', range(2, 6))
@pytest.mark.parametrize('dtype', [np.uint8, np.uint16, np.uint32])
def test_calc_bounds(ndim, dtype):
    data = np.zeros((20,) + (10,) * (ndim - 1), dtype=dtype)
    slice_arr = [slice(2, 8) for _ in range(ndim)]
    data[tuple(slice_arr)] = 1
    slice_arr[0] = slice(12, 18)
    data[tuple(slice_arr)] = 3
    low_bound, upper_bound = calc_bounds(data)
    assert low_bound.shape == (4, ndim)
    assert np.all(low_bound[0] == 25)
    assert np.all(upper_bound[0] == -1)
    assert np.all(low_bound[1] == 2)
    assert np.all(upper_bound[1] == 7)
    assert np.all(low_bound[2] == 25)
    assert np.all(upper_bound[2] == -1)
    assert np.all(low_bound[3][1:] == 2)
    assert np.all(upper_bound[3][1:] == 7)
    assert low_bound[3][0] == 12
    assert upper_bound[3][0] == 17



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, embedsignature=True
# cython: language_level=3


from cpython.mem cimport PyMem_Free
from numpy cimport float64_t, int8_t, uint8_t

from .distance_utils cimport Point, component_types

include "put_borders_in_queue.pyx"


def calculate_euclidean(np.ndarray[uint8_t, ndim=3] object_area, np.ndarray[uint8_t, ndim=3] base_object,
                        np.ndarray[int8_t, ndim=2] neighbourhood, np.ndarray[float64_t, ndim=1] distance,
                        distance_cache=None):
    """
    Calculate euclidean watersheed for one core object

    :param object_area: Area in which euclidean watershhed sholud be calculated
    :param base_object: Core object from which watersheed start
    :param neighbourhood: negihbourhood defined as array of size (neigbourhood_size, 3)
    :param distance: array for distances of negibours. Need have size (neighbourhood_szie).
        Used for handling spacinh in image.
    :return: distance from core object
    """
    cdef np.ndarray[uint8_t, ndim=3] consumed_area = np.copy(base_object)
    cdef np.ndarray[float64_t, ndim=3] result, d_cache
    cdef Size x_size, y_size, z_size, array_pos, x, y, z, xx, yy, zz
    cdef Py_ssize_t count = 0
    cdef char neigh_length = neighbourhood.shape[0]
    cdef int neigh_it
    cdef my_queue[Point] current_points
    cdef Point p, p1
    z_size = object_area.shape[0]
    y_size = object_area.shape[1]
    x_size = object_area.shape[2]
    result = np.full((z_size, y_size, x_size), np.inf, dtype=np.float64)
    result[base_object != 0] = 0
    if distance_cache is None:
        d_cache = result
    else:
        d_cache = distance_cache
    put_borders_in_queue(current_points, base_object, neighbourhood)
    while not current_points.empty():
        p = current_points.front()
        current_points.pop()
        count += 1
        """if consumed_area[p.z, p.y, p.x] > 0:
            continue"""
        consumed_area[p.z, p.y, p.x] = 0
        for neigh_it in range(neigh_length):
            z = p.z + neighbourhood[neigh_it, 0]
            y = p.y + neighbourhood[neigh_it, 1]
            x = p.x + neighbourhood[neigh_it, 2]
            if x < 0 or y < 0 or z < 0 or x >= x_size or y >= y_size or z >= z_size:
                continue
            if object_area[z, y, x] == 0:
                continue
            if d_cache[z, y, x] >= result[p.z, p.y, p.x] + distance[neigh_it]:
                if result[z, y, x] > result[p.z, p.y, p.x] + distance[neigh_it]:
                    result[z, y, x] = result[p.z, p.y, p.x] + distance[neigh_it]
                    if consumed_area[z, y, x] == 0:
                        consumed_area[z, y, x] = 1
                        p1.z = z
                        p1.y = y
                        p1.x = x
                        current_points.push(p1)
    return result



"""
Module with fast unique calculation.
This implementation is faster than numpy.unique and comparable to
pandas.unique(data.ravel())` for data with continuous memory.
If data are not continuous memory this function is faster than pandas.unique.

For benchmark please see https://github.com/4DNucleome/PartSegCore-compiled-backend/blob/master/notebooks/performance_label_unique.ipynb
"""

import numpy as np

from PartSegCore_compiled_backend._fast_unique import unique1d, unique2d, unique3d


def label_unique(array, numpy_fallback=False) -> np.ndarray:
    """
    Calculate unique values in array.

    Parameters
    ----------
    array : np.ndarray
        array to calculate unique values
    numpy_fallback : bool
        if True allow using numpy.unique if cython version is not available
        otherwise raise RuntimeError

    Returns
    -------
    np.ndarray
        array of unique values
    """
    if array.ndim == 1:
        return unique1d(array)
    if array.ndim == 2:
        return unique2d(array)
    if array.ndim == 3:
        return unique3d(array)
    if numpy_fallback:
        return np.unique(array)
    raise RuntimeError('Array must be 1d, 2d or 3d')



import numpy as np

def get_maximum_component(
    components: np.ndarray,
    data_mask: np.ndarray,
    paths: np.ndarray,
    components_translation: np.ndarray,
    num_of_components: int | None = None,
) -> np.ndarray: ...
def get_minimum_component(
    components: np.ndarray,
    data_mask: np.ndarray,
    paths: np.ndarray,
    components_translation: np.ndarray,
    num_of_components: int | None = None,
) -> np.ndarray: ...
def get_closest_component(
    components: np.ndarray,
    data_mask: np.ndarray,
    distances: np.ndarray,
    components_translation: np.ndarray,
    num_of_components: int | None = None,
) -> np.ndarray: ...



import numpy as np
import pytest

from PartSegCore_compiled_backend.fast_unique import label_unique


@pytest.mark.parametrize(
    'array',
    [
        np.random.default_rng().integers(0, 100, size=(1000,)),
        np.random.default_rng().integers(0, 100, size=(100, 100)),
        np.random.default_rng().integers(0, 100, size=(10, 100, 100)),
    ],
    ids=['1d', '2d', '3d'],
)
def test_label_unique(array):
    assert np.array_equal(np.unique(array), label_unique(array))


def test_label_unique_fallback():
    data = np.random.default_rng().integers(0, 100, size=(10, 10, 10, 10))
    with pytest.raises(RuntimeError, match='Array must be'):
        label_unique(data)
    assert np.array_equal(np.unique(data), label_unique(data, numpy_fallback=True))



# distutils: language = c++
# cython: language_level=3

from libcpp cimport bool
from libcpp.vector cimport vector
from numpy cimport float32_t, float64_t, int8_t, int16_t, uint8_t, uint16_t, uint32_t

import numpy

cimport numpy

import cython

ctypedef fused image_types:
    int8_t
    int16_t
    uint8_t
    uint16_t
    uint32_t
    float32_t
    float64_t

ctypedef double mu_type
ctypedef uint8_t component_type
np_component_type = numpy.uint8


cpdef enum MuType:
    base_mu = 1
    reflection_mu = 2
    two_object_mu = 3

cpdef enum SegmentType:
    separate_objects = 1
    separate_from_background = 2


cdef extern from 'mso.h' namespace 'MSO':
    cdef cppclass MSO[T, M]:
        void set_neighbourhood(vector[int8_t] neighbourhood, vector[double] distances) except +
        void erase_data()
        void set_mu_copy(const vector[M] & mu) except +
        void set_mu_copy(M * mu, size_t length) except +
        void set_mu_swap(vector[M] & mu) except +
        void set_neighbourhood(vector[int8_t] neighbourhood, vector[mu_type] distances) except +
        void set_neighbourhood(int8_t * neighbourhood, M * distances, size_t neigh_size) except +
        void set_components_num(T components_num)
        void compute_FDT(vector[M] & array) except +
        void set_use_background(bool val)
        size_t optimum_erosion_calculate(vector[M] &fdt_array, vector[T] &components_arr, vector[bool] & sprawl_area) except +
        size_t constrained_dilation(vector[M] &fdt_array, vector[T] &components_arr, vector[bool] & sprawl_area) except +
        size_t get_length()
        size_t run_MSO() except + nogil
        size_t run_MSO(size_t steps_limits) except + nogil
        size_t run_MSO(size_t steps_limits, size_t count_steps_factor) except + nogil
        void set_data[W](T * components, W size, T background_component)
        void set_data[W](T * components, W size)
        size_t steps_done()
        vector[mu_type] get_fdt()
        vector[T] get_result_catted()

    cdef cppclass MuCalc[R, T]:
        vector[R] calculate_mu_array(T *array, size_t length, T lower_bound, T upper_bound)
        vector[R] calculate_mu_array_masked(T *array, size_t length, T lower_bound, T upper_bound, uint8_t *mask)
        vector[R] calculate_reflection_mu_array(T *array, size_t length, T lower_bound, T upper_bound)
        vector[R] calculate_reflection_mu_array_masked(T *array, size_t length, T lower_bound, T upper_bound, uint8_t *mask)
        vector[R] calculate_two_object_mu(T *array, size_t length, T lower_bound, T upper_bound,
                                                  T lower_mid_bound, T upper_mid_bound)
        vector[R] calculate_two_object_mu_masked(T *array, size_t length, T lower_bound, T upper_bound,
                                                     T lower_mid_bound, T upper_mid_bound, uint8_t *mask)

cdef vector[mu_type] calculate_mu_vector(numpy.ndarray[image_types, ndim=3] image, image_types lower_bound,
                                         image_types upper_bound, MuType type_, mask=None,
                                         image_types lower_mid_bound=0, image_types upper_mid_bound=0):
    cdef vector[mu_type] result;
    cdef MuCalc[mu_type, image_types] calc
    cdef numpy.ndarray[uint8_t, ndim=3] mask_array
    cdef size_t length, x
    length = image.size

    if lower_bound > upper_bound:
        reflected = True
        tmp = lower_bound
        lower_bound = upper_bound
        upper_bound = tmp
    else:
        reflected = False
    if mask is not None:
        mask_array = mask.astype(numpy.uint8)
    if type_ == MuType.base_mu:
        if mask is None:
            result = calc.calculate_mu_array(<image_types *> image.data, length, lower_bound, upper_bound)
        else:
            result = calc.calculate_mu_array_masked(<image_types *> image.data, length, lower_bound, upper_bound,
                                               <uint8_t *> mask_array.data)
    elif type_ == MuType.reflection_mu:
        if mask is None:
            result = calc.calculate_reflection_mu_array(<image_types *> image.data, length, lower_bound, upper_bound)
        else:
            result = calc.calculate_reflection_mu_array_masked(<image_types *> image.data, length, lower_bound, upper_bound,
                                                          <uint8_t *> mask_array.data)
        reflected = False
    elif type_ == MuType.two_object_mu:
        if mask is None:
            result = calc.calculate_two_object_mu(<image_types *> image.data, length, lower_bound, upper_bound,
                                             lower_mid_bound, upper_mid_bound)
        else:
            result = calc.calculate_two_object_mu_masked(<image_types *> image.data, length, lower_bound, upper_bound,
                                                    lower_mid_bound, upper_mid_bound,
                                                    <uint8_t *> mask_array.data)
    if reflected:
        if mask is None:
            for x in range(length):
                result[x] = 1 - result[x]
        else:
            for x in range(length):
                if mask_array.data[x]:
                    result[x] = 1 - result[x]
    return result


def calculate_mu(numpy.ndarray[image_types, ndim=3] image, image_types lower_bound,
                 image_types upper_bound, MuType type_, mask=None, image_types lower_mid_bound=0,
                 image_types upper_mid_bound=0):
    cdef vector[mu_type] tmp;
    tmp = calculate_mu_vector(image, lower_bound, upper_bound, type_, mask, lower_bound, upper_mid_bound)
    result = numpy.array(tmp)
    return result.reshape((image.shape[0], image.shape[1], image.shape[2]))


cdef class PyMSO:
    cdef MSO[component_type, mu_type] mso
    cdef numpy.ndarray components
    cdef component_type components_num

    def set_image(self, numpy.ndarray[image_types, ndim=3] image, image_types lower_bound,
                  image_types upper_bound, MuType type_, mask=None,
                  image_types lower_mid_bound=0, image_types upper_mid_bound=0):
        cdef vector[mu_type] mu
        mu = calculate_mu_vector(image, lower_bound, upper_bound, type_, mask, lower_mid_bound, upper_mid_bound)
        self.mso.set_mu_swap(mu)

    def set_mu_array(self, numpy.ndarray[mu_type, ndim=3] mu):
        self.mso.set_mu_copy(<mu_type *> mu.data, mu.size)

    def set_components(self, numpy.ndarray[component_type, ndim=3] components, component_num=None):
        cdef vector[uint16_t] shape = vector[uint16_t](3, 0)
        shape[0] = components.shape[0]
        shape[1] = components.shape[1]
        shape[2] = components.shape[2]
        self.mso.set_data(<component_type *> components.data, shape)
        if component_num is not None:
            self.mso.set_components_num(component_num+2)
        else:
            self.mso.set_components_num(components.max()+2)
        self.components = components

    def set_neighbourhood(self, numpy.ndarray[int8_t, ndim=2] neighbourhood, numpy.ndarray[mu_type] distances):
        self.mso.set_neighbourhood(<int8_t *> neighbourhood.data, <mu_type *> distances.data, distances.size)

    def calculate_FDT(self):
        cdef vector[mu_type] fdt = vector[mu_type](self.mso.get_length(), 0)
        self.mso.compute_FDT(fdt)
        return numpy.array(fdt).reshape(self.components.shape[0], self.components.shape[1], self.components.shape[2])

    def optimum_erosion_calculate(self, numpy.ndarray[mu_type, ndim=3] fdt_array,
                           numpy.ndarray[component_type, ndim=3] components_arr,
                           numpy.ndarray[uint8_t, ndim=3] sprawl_area):
        """ For testing purpose """
        cdef vector[mu_type] fdt
        fdt.assign(<mu_type *> fdt_array.data, (<mu_type *> fdt_array.data) + (<size_t> fdt_array.size))
        cdef vector[component_type] components
        components.assign(<component_type *> components_arr.data,
                              (<component_type *> components_arr.data) + (<size_t> components_arr.size))
        cdef vector[bool] sprawl_vec
        sprawl_vec.resize(sprawl_area.size)
        cdef numpy.ndarray[uint8_t, ndim=1] tmp = sprawl_area.flatten()
        cdef Py_ssize_t x
        for x in range(tmp.size):
            sprawl_vec[x] = <bool> tmp[x]

        self.mso.set_components_num(components_arr.max())

        count = self.mso.optimum_erosion_calculate(fdt, components, sprawl_vec)
        res = numpy.array(components, dtype=np_component_type)
        res = res.reshape([components_arr.shape[i] for i in range(components_arr.ndim)])
        return res

    def constrained_dilation(self, numpy.ndarray[mu_type, ndim=3] fdt_array,
                           numpy.ndarray[component_type, ndim=3] components_arr,
                           numpy.ndarray[uint8_t, ndim=3] sprawl_area):
        """ For testing purpose """
        cdef vector[mu_type] fdt
        fdt.assign(<mu_type *> fdt_array.data, (<mu_type *> fdt_array.data) + (<size_t> fdt_array.size))
        cdef vector[component_type] components
        components.assign(<component_type *> components_arr.data,
                              (<component_type *> components_arr.data) + (<size_t> components_arr.size))
        cdef vector[bool] sprawl_vec
        sprawl_vec.resize(sprawl_area.size)
        cdef numpy.ndarray[uint8_t, ndim=1] tmp = sprawl_area.flatten()
        cdef Py_ssize_t x
        for x in range(tmp.size):
            sprawl_vec[x] = <bool> tmp[x]

        self.mso.set_components_num(components_arr.max())

        count = self.mso.constrained_dilation(fdt, components, sprawl_vec)
        res = numpy.array(components, dtype=np_component_type)
        res = res.reshape([components_arr.shape[i] for i in range(components_arr.ndim)])
        return res

    def run_MSO(self, size_t step_limits=1, size_t count_steps_factor=3):
        cdef size_t val
        with nogil:
            val = self.mso.run_MSO(step_limits, count_steps_factor)
        return val

    def steps_done(self):
        return self.mso.steps_done()

    def set_components_num(self, num):
        self.mso.set_components_num(num+2)

    def get_result_catted(self):
        cdef vector[component_type] res = self.mso.get_result_catted()
        res_arr = numpy.array(res, dtype=np_component_type)
        res_arr = res_arr.reshape([self.components.shape[i] for i in range(self.components.ndim)])
        return res_arr

    def get_fdt(self):
        cdef vector[mu_type] res = self.mso.get_fdt()
        res_arr = numpy.array(res, dtype=numpy.float64)
        res_arr = res_arr.reshape([self.components.shape[i] for i in range(self.components.ndim)])
        return res_arr

    def set_use_background(self, use):
        self.mso.set_use_background(use)


@cython.boundscheck(False)
def calculate_mu_mid(numpy.ndarray[image_types, ndim=3] image, image_types lower_bound, image_types mid_point,
                 image_types upper_bound):
    cdef int16_t x_size, y_size, z_size, x, y, z
    cdef mu_type res_val
    cdef image_types point_val
    x_size = image.shape[2]
    y_size = image.shape[1]
    z_size = image.shape[0]
    cdef numpy.ndarray[mu_type, ndim=3] res = numpy.empty((z_size, y_size, x_size), dtype=numpy.float64)
    if upper_bound < lower_bound:
        tmp = lower_bound
        lower_bound = upper_bound
        upper_bound = tmp

    for z in range(z_size):
        for y in range(y_size):
            for x in range(x_size):
                point_val = image[z, y, x]
                if point_val < mid_point:
                    res_val = (mid_point - point_val)/(mid_point - lower_bound)
                else:
                    res_val = (point_val - mid_point)/(upper_bound - mid_point)
                if res_val > 1:
                    res_val = 1
                elif res_val < 0:
                    res_val = 0
                res[z, y, x] = res_val
    return res



"""
This module contains function to perform multiple component sprawl (watershed like)
"""

from __future__ import annotations

import os
import typing
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial

import numpy as np

from PartSegCore_compiled_backend.sprawl_utils.euclidean_cython import calculate_euclidean
from PartSegCore_compiled_backend.sprawl_utils.fuzzy_distance import MuType, calculate_mu_array, fuzzy_distance
from PartSegCore_compiled_backend.sprawl_utils.path_sprawl_cython import calculate_maximum, calculate_minimum
from PartSegCore_compiled_backend.sprawl_utils.sprawl_utils import (
    get_closest_component,
    get_maximum_component,
    get_minimum_component,
)


def _allocate_cache(distance_cache: np.ndarray | None, data_shape: tuple):
    if distance_cache is None or distance_cache.shape[0] < 3:
        cache_size = 5
        distance_cache = np.zeros((cache_size + 1, *data_shape), dtype=np.float64)
    else:
        cache_size = distance_cache.shape[0] - 1
    components_numbers_translate = np.zeros(cache_size + 1, dtype=np.uint32)
    return distance_cache, cache_size, components_numbers_translate


def path_maximum_sprawl(
    data_f: np.ndarray,
    components: np.ndarray,
    components_count: int,
    neighbourhood: np.ndarray,
    distance_cache: np.ndarray | None = None,
    data_cache: np.ndarray | None = None,
) -> np.ndarray:
    """
    Calculate sprawl in respect to brightness. Distance between voxels is the maximum brightness on
    all paths connecting them.

    Parameters
    ----------
    data_f : np.ndarray
        Array with brightness.
    components : np.ndarray
        Core components as labeled array.
    components_count : int
        Number of components.
    neighbourhood : np.ndarray
        Information about neighbourhood, shift array.
    distance_cache : np.ndarray, optional
        Cache array for reducing memory allocation. Shape of data_m, dtype is np.float64.
    data_cache : np.ndarray, optional
        Cache array for reducing memory allocation. Shape of data_m, dtype is data_m.dtype
        (use np.uint8 instead of np.bool)

    Returns
    -------
    np.ndarray
        Array with updated labels.
    """

    if data_cache is None:
        data_cache = np.zeros(data_f.shape, data_f.dtype)
    if components_count == 1:
        np.copyto(data_cache, data_f)
        tmp = calculate_maximum(data_cache, (components == 1).astype(np.uint8), neighbourhood)
        components[tmp > 0] = 1
        return components
    base_components = np.copy(components)
    masked_area = ((data_f > 0) * (base_components == 0)).astype(np.uint8)
    distance_cache, cache_size, components_numbers_translate = _allocate_cache(distance_cache, data_f.shape)
    distance_cache[:] = -np.inf
    current_in_cache = 0
    for component in range(1, components_count + 1):
        np.copyto(data_cache, data_f)
        data_cache[(base_components > 0) * (base_components != component)] = 0
        current_in_cache += 1
        distance_cache[current_in_cache] = calculate_maximum(
            data_cache, (components == component).astype(np.uint8), neighbourhood
        )
        components_numbers_translate[current_in_cache] = component
        if current_in_cache == cache_size:
            components = get_maximum_component(
                components, masked_area, distance_cache, components_numbers_translate, current_in_cache
            )
            current_in_cache = 0
    if current_in_cache > 0:
        components = get_maximum_component(
            components, masked_area, distance_cache, components_numbers_translate, current_in_cache
        )
    return components


def path_minimum_sprawl(
    data_f: np.ndarray,
    components: np.ndarray,
    components_count: int,
    neighbourhood: np.ndarray,
    distance_cache: np.ndarray | None = None,
    data_cache: np.ndarray | None = None,
) -> np.ndarray:
    """
    Calculate sprawl in respect to brightness. Distance between voxels is the maximum brightness on
    all paths connecting them.

    Parameters
    ----------
    data_f : np.ndarray
        Array with brightness.
    components : np.ndarray
        Core components as labeled array.
    components_count : int
        Number of components.
    neighbourhood : np.ndarray
        Information about neighbourhood, shift array.
    distance_cache : np.ndarray, optional
        Cache array for reducing memory allocation. Shape of data_m, dtype is np.float64.
    data_cache : np.ndarray, optional
        Cache array for reducing memory allocation. Shape of data_m, dtype is data_m.dtype (use np.uint8 instead of np.bool).

    Returns
    -------
    np.ndarray
        Array with updated labels.
    """
    maximum = data_f.max()
    if data_cache is None:
        data_cache = np.zeros(data_f.shape, data_f.dtype)
    if components_count == 1:
        np.copyto(data_cache, data_f)
        tmp = calculate_minimum(data_cache, (components == 1).astype(np.uint8), neighbourhood)
        components[tmp < maximum] = 1
        return components
    base_components = np.copy(components)
    masked_area = ((data_f > 0) * (base_components == 0)).astype(np.uint8)
    distance_cache, cache_size, components_numbers_translate = _allocate_cache(distance_cache, data_f.shape)
    distance_cache[:] = np.inf
    current_in_cache = 0
    for component in range(1, components_count + 1):
        np.copyto(data_cache, data_f)
        data_cache[(base_components > 0) * (base_components != component)] = 0
        current_in_cache += 1
        distance_cache[current_in_cache] = calculate_minimum(
            data_cache, (components == component).astype(np.uint8), neighbourhood
        )
        components_numbers_translate[current_in_cache] = component
        if current_in_cache == cache_size:
            components = get_minimum_component(
                components, masked_area, distance_cache, components_numbers_translate, current_in_cache
            )
            current_in_cache = 0
    if current_in_cache > 0:
        components = get_minimum_component(
            components, masked_area, distance_cache, components_numbers_translate, current_in_cache
        )
    return components


def euclidean_sprawl(
    data_m: np.ndarray,
    components: np.ndarray,
    components_count: int,
    neigh_arr,
    dist_arr,
    distance_cache=None,
    data_cache=None,
) -> np.ndarray:
    """
    Calculate euclidean sprawl (watershed)

    Parameters
    ----------
    data_m : np.ndarray
        Area for sprawl.
    components : np.ndarray
        Core components as labeled array.
    components_count : int
        Number of components.
    neigh_arr : object
        Information about neighbourhood, shift array.
    dist_arr : object
        Information about neighbourhood, distance array for shifted position.
    distance_cache : np.ndarray, optional
        Cache array for reducing memory allocation. Shape of data_m, dtype is np.float64.
    data_cache : np.ndarray, optional
        Cache array for reducing memory allocation. Shape of data_m, dtype is data_m.dtype
        (use np.uint8 instead of np.bool).

    Returns
    -------
    np.ndarray
        Array with updated labels.
    """
    #  return calculate_euclidean_iterative(data_m, components, neigh_arr, dist_arr, components_count)
    return distance_sprawl(
        calculate_euclidean, data_m, components, components_count, neigh_arr, dist_arr, distance_cache, data_cache
    )


def fdt_sprawl(
    data_m: np.ndarray,
    components: np.ndarray,
    components_count: int,
    neigh_arr,
    dist_arr,
    lower_bound,
    upper_bound,
    distance_cache=None,
    data_cache=None,
) -> np.ndarray:
    """
    Calculate the FDT (Fuzzy Distance Transform) sprawl.

    Parameters
    ----------
    data_m : np.ndarray
        The sprawl area.
    components : np.ndarray
        Core components as labeled array.
    components_count : int
        Number of components.
    neigh_arr : array-like
        Information about the neighborhood, shift array.
    dist_arr : array-like
        Information about the neighborhood, distance array for shifted position.
    lower_bound : float
        Lower bound for calculating the mu value.
    upper_bound : float
        Upper bound for calculating the mu value.
    distance_cache : np.ndarray, optional
        Cache array for reducing memory allocation. Shape of data_m, dtype is np.float64.
    data_cache : np.ndarray, optional
        Cache array for reducing memory allocation. Shape of data_m, dtype is data_m.dtype
        (use np.uint8 instead of np.bool).

    Returns
    -------
    np.ndarray
        Array with updated labels.
    """
    if lower_bound > upper_bound:
        mu_array = 1 - calculate_mu_array(data_m, upper_bound, lower_bound, MuType.reflection_mu)
    else:
        mu_array = calculate_mu_array(data_m, lower_bound, upper_bound, MuType.reflection_mu)
    return distance_sprawl(
        partial(fuzzy_distance, mu_array=mu_array),
        data_m,
        components,
        components_count,
        neigh_arr,
        dist_arr,
        distance_cache,
        data_cache,
    )


def sprawl_component(
    data_m: np.ndarray,
    components: np.ndarray,
    component_number: int,
    neigh_arr: np.ndarray,
    dist_arr: np.ndarray,
    calculate_operator: typing.Callable,
) -> np.ndarray:
    """
    Calculate the sprawl for a single component.

    Parameters
    ----------
    data_m : np.ndarray
        The sprawl area.
    components : np.ndarray
        Array of components.
    component_number : int
        Chosen component number.
    neigh_arr : np.ndarray
        Information about the neighborhood, shift array.
    dist_arr : np.ndarray
        Information about the neighborhood, distance array for shifted position.
    calculate_operator : typing.Callable
        Function to be called to calculate sprawl.

    Returns
    -------
    np.ndarray
        Array with updated labels.
    """
    data_cache = np.copy(data_m)
    data_cache[(components > 0) * (components != component_number)] = 0
    return calculate_operator(
        data_cache, np.array(components == component_number).astype(np.uint8), neigh_arr, dist_arr
    )


def distance_sprawl(
    calculate_operator,
    data_m: np.ndarray,
    components: np.ndarray,
    components_count: int,
    neigh_arr,
    dist_array,
    distance_cache=None,
    data_cache=None,
    parallel=False,
) -> np.ndarray:
    if data_m.dtype == bool:
        data_m = data_m.astype(np.uint8)
    if data_cache is None:
        data_cache = np.zeros(data_m.shape, data_m.dtype)
    if components_count == 1:
        np.copyto(data_cache, data_m)
        tmp = calculate_operator(data_cache, (components == 1).astype(np.uint8), neigh_arr, dist_array)
        components[tmp < 2**17] = 1
        return components
    base_components = np.copy(components)
    masked_area = ((data_m > 0) * (base_components == 0)).astype(np.uint8)
    distance_cache, cache_size, components_numbers_translate = _allocate_cache(distance_cache, data_m.shape)
    distance_cache[:] = np.inf
    current_in_cache = 0
    if parallel:
        workers_num = os.cpu_count()
        if not workers_num:
            workers_num = 4
        with ThreadPoolExecutor(max_workers=workers_num) as executor:
            result_info = {
                executor.submit(
                    sprawl_component,
                    data_m,
                    base_components,
                    component_number,
                    neigh_arr,
                    dist_array,
                    calculate_operator,
                ): component_number
                for component_number in range(1, components_count + 1)
            }
            for result in as_completed(result_info):
                component_number = result_info[result]
                current_in_cache += 1
                components_numbers_translate[current_in_cache] = component_number
                distance_cache[current_in_cache] = result.result()
                if current_in_cache == cache_size:
                    components = get_closest_component(
                        components, masked_area, distance_cache, components_numbers_translate, current_in_cache
                    )
                    current_in_cache = 0

    else:
        for component in range(1, components_count + 1):
            np.copyto(data_cache, data_m)
            data_cache[(base_components > 0) * (base_components != component)] = 0
            current_in_cache += 1
            distance_cache[current_in_cache] = calculate_operator(
                data_cache,
                (components == component).astype(np.uint8),
                neigh_arr,
                dist_array,
                distance_cache=distance_cache[0],
            )
            components_numbers_translate[current_in_cache] = component
            if current_in_cache == cache_size:
                components = get_closest_component(
                    components, masked_area, distance_cache, components_numbers_translate, current_in_cache
                )
                current_in_cache = 0
    if current_in_cache > 0:
        components = get_closest_component(
            components, masked_area, distance_cache, components_numbers_translate, current_in_cache
        )
    return components


def reverse_permutation(perm: list[int]) -> list[int]:
    rev = [0] * len(perm)
    for i, x in enumerate(perm, 1):
        rev[x - 1] = i
    return rev


def relabel_with_perm(labeling: list, perm: list) -> list:
    perm = reverse_permutation(perm)
    return [perm[x] for x in labeling]


def verify_cohesion(elements: list[int], graph: list[list[int]]) -> bool:
    start = elements[0]
    elements_set = set(elements)
    elements_set.remove(start)
    queue = list(graph[start])
    while queue and elements_set:
        el = queue.pop()
        if el in elements_set:
            elements_set.remove(el)
            queue.extend(graph[el])
    return len(elements_set) == 0


def relabel_array(data: np.ndarray, perm: list[int]) -> np.ndarray:
    result = np.zeros(data.shape, dtype=data.dtype)
    for i, val in enumerate(perm):
        result[data == i + 1] = val
    return result



from itertools import product

import numpy as np
import pytest

from PartSegCore_compiled_backend.color_image_cython import calculate_borders, calculate_borders2d


@pytest.mark.parametrize('label_type', [np.uint8, np.uint16, np.uint32])
def test_calculate_borders(label_type):
    layers = np.zeros((1, 10, 10, 10), dtype=label_type)
    layers[:, 3:-3, 3:-3, 3:-3] = 1
    res = calculate_borders(layers, 0, False)
    expected = layers.copy()
    expected[:, 4:-4, 4:-4, 4:-4] = 0
    assert np.all(res == expected)
    res = calculate_borders(layers, 0, True)
    expected = layers.copy()
    expected[:, :, 4:-4, 4:-4] = 0
    assert np.all(res == expected)
    res = calculate_borders2d(layers, 0)
    assert np.all(res == expected)


@pytest.mark.parametrize('label_type', [np.uint8, np.uint16, np.uint32])
def test_calculate_borders_thick(label_type):
    layers = np.zeros((1, 16, 16, 16), dtype=label_type)
    layers[:, 3:-3, 3:-3, 3:-3] = 1
    res = calculate_borders(layers, 1, False)
    expected = np.zeros((1, 16, 16, 16), dtype=np.uint8)
    expected[:, 2:-2, 2:-2, 2:-2] = 1
    expected[:, 5:-5, 5:-5, 5:-5] = 0
    for c1, c2 in product([2, -3], repeat=2):
        for x in range(3):
            cord = [c1, c2]
            cord.insert(x, slice(None))
            expected[(0, *cord)] = 0

    assert np.all(res == expected)


@pytest.mark.parametrize('label_type', [np.uint8, np.uint16, np.uint32])
def test_calculate_borders_thick2d(label_type):
    layers = np.zeros((1, 16, 16, 16), dtype=label_type)
    layers[:, 3:-3, 3:-3, 3:-3] = 1
    res1 = calculate_borders(layers, 1, True)
    res2 = calculate_borders2d(layers, 1)
    assert np.all(res1 == res2)
    expected = np.zeros((1, 16, 16, 16), dtype=np.uint8)
    expected[:, 3:-3, 2:-2, 2:-2] = 1
    expected[:, :, 5:-5, 5:-5] = 0
    for c1, c2 in product([2, -3], repeat=2):
        expected[0, :, c1, c2] = 0

    assert np.all(res1 == expected)
    assert np.all(res2 == expected)



# distutils: define_macros=CYTHON_TRACE_NOGIL=1
# distutils: language = c++
# cython: language_level=3, boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, embedsignature=True

import numpy as np

cimport numpy as np
cimport openmp

from cython.parallel import prange, parallel, threadid
from cython.operator cimport dereference as deref, preincrement as inc

from libcpp.unordered_set cimport unordered_set
from libcpp.vector cimport vector


ctypedef fused numpy_int_types:
    np.uint8_t
    np.uint16_t
    np.uint32_t
    np.int8_t
    np.int16_t
    np.int32_t
    np.int64_t


def unique1d(np.ndarray[numpy_int_types, ndim=1] array):
    cdef unordered_set[numpy_int_types] unique_values
    cdef vector[unordered_set[numpy_int_types]] unique_values_vector
    cdef Py_ssize_t i, j, n_rows, n_cols, id_num, n_items
    cdef numpy_int_types prev_val, v
    n_items = array.shape[0]
    unique_values_vector.resize(50) # openmp.omp_get_num_threads())

    with nogil, parallel():
        id_num = threadid()

        for i in prange(n_items):
            prev_val = array[i]
            unique_values_vector[id_num].insert(prev_val)
            if array[i] != prev_val:
                prev_val = array[i]
                unique_values_vector[id_num].insert(prev_val)

    for s in unique_values_vector:
        for v in s:
            unique_values.insert(v)
    return np.sort(np.fromiter(unique_values, dtype=array.dtype))


def unique2d(np.ndarray[numpy_int_types, ndim=2] array):
    cdef unordered_set[numpy_int_types] unique_values
    cdef vector[unordered_set[numpy_int_types]] unique_values_vector
    cdef Py_ssize_t i, j, n_rows, n_cols, id_num
    cdef numpy_int_types prev_val, v
    n_rows = array.shape[0]
    n_cols = array.shape[1]
    unique_values_vector.resize(50) # openmp.omp_get_num_threads())

    with nogil, parallel():
        id_num = threadid()

        for i in prange(n_rows):
            prev_val = array[i, 0]
            unique_values_vector[id_num].insert(prev_val)
            for j in range(n_cols):
                if array[i, j] != prev_val:
                    prev_val = array[i, j]
                    unique_values_vector[id_num].insert(prev_val)

    for s in unique_values_vector:
        for v in s:
            unique_values.insert(v)
    return np.sort(np.fromiter(unique_values, dtype=array.dtype))


def unique3d(np.ndarray[numpy_int_types, ndim=3] array):
    cdef unordered_set[numpy_int_types] unique_values
    cdef unordered_set[numpy_int_types] unique_values_local
    cdef vector[unordered_set[numpy_int_types]] unique_values_vector
    cdef Py_ssize_t i, j, k, n_rows, n_cols, n_depth, id_num
    cdef numpy_int_types prev_val, v
    cdef unordered_set[numpy_int_types].iterator begin_it, end_it
    n_rows = array.shape[0]
    n_cols = array.shape[1]
    n_depth = array.shape[2]

    unique_values_vector.resize(50)  # openmp.omp_get_num_threads())

    with nogil, parallel():
        id_num = threadid()


        for i in prange(n_rows):
            prev_val = array[i, 0, 0]
            unique_values_vector[id_num].insert(prev_val)
            for j in range(n_cols):
                for k in range(n_depth):
                    if array[i, j, k] != prev_val:
                        prev_val = array[i, j, k]
                        unique_values_vector[id_num].insert(prev_val)

    for s in unique_values_vector:
        for v in s:
            unique_values.insert(v)

    return np.sort(np.fromiter(unique_values, dtype=array.dtype))



import numpy as np

def calculate_maximum(
    object_area: np.ndarray, base_object: np.ndarray, neighbourhood: np.ndarray, result: np.ndarray | None = None
) -> np.ndarray: ...
def calculate_minimum(
    object_area: np.ndarray, base_object: np.ndarray, neighbourhood: np.ndarray, result: np.ndarray | None = None
) -> np.ndarray: ...



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True
# cython: language_level=3

from numpy cimport float64_t, int8_t, uint8_t

from .distance_utils cimport Point

include "put_borders_in_queue.pyx"


def calculate_maximum(np.ndarray[float64_t, ndim=3] object_area, np.ndarray[uint8_t, ndim=3] base_object,
                      np.ndarray[int8_t, ndim=2] neighbourhood, result=None):
    """
    Calculate maximum path from source

    :param object_area: data with removed other components
    :param base_object: area of component
    :param neighbourhood: array with relative coordinates of neighbours
    :param result: cache array to store result. If calling function multiple times allows to skip memory allocation.
    :return:
    """
    if result is None:
        result = np.full((object_area.shape[0], object_area.shape[1], object_area.shape[2]), -np.inf, dtype=np.float64)
    res, c = _calculate_maximum(object_area, base_object, neighbourhood, result)
    return res


cdef _calculate_maximum(np.ndarray[float64_t, ndim=3] object_area, np.ndarray[uint8_t, ndim=3] base_object,
                       np.ndarray[int8_t, ndim=2] neighbourhood, np.ndarray[float64_t, ndim=3] result):
    # cdef np.ndarray[np.uint8_t, ndim=3] consumed_area = np.copy(base_object)
    # cdef np.ndarray[np.float64_t, ndim=3] result
    cdef Size x_size, y_size, z_size, array_pos, x, y, z, xx, yy, zz
    cdef Py_ssize_t count = 0
    cdef char neigh_length = neighbourhood.shape[0]
    cdef int neigh_it
    cdef my_queue[Point] current_points
    cdef Point p, p1
    cdef np.float64_t object_area_value, result_value
    z_size = object_area.shape[0]
    y_size = object_area.shape[1]
    x_size = object_area.shape[2]


    result[base_object > 0] = object_area[base_object > 0]

    # Find border of component
    put_borders_in_queue(current_points, base_object, neighbourhood)
    while not current_points.empty():
        count+=1
        p = current_points.front()
        current_points.pop()
        base_object[p.z, p.y, p.x] = 0
        for neigh_it in range(neigh_length):
            z = p.z + neighbourhood[neigh_it, 0]
            y = p.y + neighbourhood[neigh_it, 1]
            x = p.x + neighbourhood[neigh_it, 2]
            if x < 0 or y < 0 or z < 0 or x >= x_size or y >= y_size or z >= z_size:
                continue
            object_area_value = object_area[z, y, x]
            if object_area_value == 0:
                continue
            if result[z, y, x] == object_area_value:
                continue
            if result[p.z, p.y, p.x] > result[z, y, x]:
                if result[p.z, p.y, p.x] > object_area_value:
                    result[z, y, x] = object_area_value
                else:
                    result[z, y, x] = result[p.z, p.y, p.x]
                if base_object[z, y, x] == 0:
                    p1.z = z
                    p1.y = y
                    p1.x = x
                    current_points.push(p1)
                    base_object[z, y, x] = 1
    return result, count

def calculate_minimum(np.ndarray[float64_t, ndim=3] object_area, np.ndarray[uint8_t, ndim=3] base_object,
                      np.ndarray[int8_t, ndim=2] neighbourhood, result=None):
    """
    Calculate maximum path from source

    :param maximum: maximum possible value on path
    :param object_area: data with removed other components
    :param base_object: area of component
    :param neighbourhood: array with relative coordinates of neighbours
    :param result:
    :return:
    """
    if result is None:
        result = np.full((object_area.shape[0], object_area.shape[1], object_area.shape[2]), np.inf, dtype=np.float64)
    res, c = _calculate_minimum(object_area, base_object, neighbourhood, result)
    return res

cdef _calculate_minimum(np.ndarray[float64_t, ndim=3] object_area, np.ndarray[uint8_t, ndim=3] base_object,
                       np.ndarray[int8_t, ndim=2] neighbourhood, np.ndarray[float64_t, ndim=3] result):
    # cdef np.ndarray[np.uint8_t, ndim=3] consumed_area = np.copy(base_object)
    # cdef np.ndarray[np.float64_t, ndim=3] result
    cdef Size x_size, y_size, z_size, array_pos, x, y, z, xx, yy, zz
    cdef Py_ssize_t count = 0
    cdef char neigh_length = neighbourhood.shape[0]
    cdef int neigh_it
    cdef my_queue[Point] current_points
    cdef Point p, p1
    cdef np.float64_t object_area_value, result_value
    z_size = object_area.shape[0]
    y_size = object_area.shape[1]
    x_size = object_area.shape[2]

    result[base_object > 0] = object_area[base_object > 0]
    put_borders_in_queue(current_points, base_object, neighbourhood)
    while not current_points.empty():
        count+=1
        p = current_points.front()
        current_points.pop()
        base_object[p.z, p.y, p.x] = 0
        for neigh_it in range(neigh_length):
            z = p.z + neighbourhood[neigh_it, 0]
            y = p.y + neighbourhood[neigh_it, 1]
            x = p.x + neighbourhood[neigh_it, 2]
            if x < 0 or y < 0 or z < 0 or x >= x_size or y >= y_size or z >= z_size:
                continue
            object_area_value = object_area[z, y, x]
            if object_area_value == 0:
                continue
            if result[z, y, x] == object_area_value:
                continue
            if result[p.z, p.y, p.x] < result[z, y, x]:
                if result[p.z, p.y, p.x] < object_area_value:
                    result[z, y, x] = object_area_value
                else:
                    result[z, y, x] = result[p.z, p.y, p.x]
                if base_object[z, y, x] == 0:
                    p1.z = z
                    p1.y = y
                    p1.x = x
                    current_points.push(p1)
                    base_object[z, y, x] = 1
    return result, count
